# Copyright contributors to the TSFM project
#
"""Utilities to support conformal forecasts"""

import enum
import json
import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from scipy.stats import norm
from transformers.feature_extraction_utils import PreTrainedFeatureExtractor

from tsfm_public.toolkit.processor import BaseProcessor


LOGGER = logging.getLogger(__file__)


class NonconformityScores(enum.Enum):
    """`Enum` for the different kinds of nonconformity scores."""

    ABSOLUTE_ERROR = "absolute_error"
    ERROR = "error"


class PositiveNonconformityScores(enum.Enum):
    """`Enum` for the different kinds of positive nonconformity scores."""

    ABSOLUTE_ERROR = "absolute_error"


class ThresholdFunction(enum.Enum):
    """`Enum` for the different kinds of nonconformity score thresholding functions."""

    WEIGHTING = "weighting"


class Weighting(enum.Enum):
    """`Enum` for the different kinds of nonconformity score weighting approaches."""

    UNIFORM = "uniform"
    EXPONENTIAL_DECAY = "exponential_decay"


class WeightingOptimization(enum.Enum):
    """`Enum` for the different nonconformity score weighting optimization approaches."""

    WASS1 = "wass1"


class PostHocProbabilisticMethod(enum.Enum):
    CONFORMAL = "conformal"
    GAUSSIAN = "gaussian"


"""
Post-Hoc Probabilistic Wrapper Classes
"""


class PostHocProbabilisticProcessor(BaseProcessor):
    """Entry point for posthoc probabilistic approaches for forecasting

    We adopt HF FeatureExtractionMixin through the BaseProcessor class to create the processor and support
    serialization and deserialization

    """

    PROCESSOR_NAME = "conformal_config.json"

    def __init__(
        self,
        window_size: Optional[int] = None,
        quantiles: List[float] = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
        nonconformity_score: str = NonconformityScores.ABSOLUTE_ERROR.value,
        method: str = PostHocProbabilisticMethod.CONFORMAL.value,
        weighting: str = Weighting.UNIFORM.value,
        weighting_params: Dict[str, Any] = {},
        threshold_function: str = ThresholdFunction.WEIGHTING.value,
        aggregation: Union[str, int] = "median",
        aggregation_axis: Union[int, Tuple[int, ...]] = 1,
        **kwargs,
    ):
        """
        PostHoc Probabilistic Processor. Turns the point estimates of a multivariate forecast model into probabilistic
        forecasts (quantile estimate of the target variable)

        Args:
            window_size (int, optional): Maximum context window size for considering past residuals. If None
                (default), all previous provided observations are used.
            quantiles (List[float]): List of target quantiles to compute, with values in the open interval (0, 1).
            nonconformity_score (str, optional): Name of the nonconformity score to use, as defined in the
                `NonconformityScores` enum. Applicable only if the method is conformal.
            method (str): Name of the post-hoc probabilistic method to use, as defined in the
                `PostHocProbabilisticMethod` enum.
            weighting (str): Strategy for weighting nonconformity scores, as defined in the `Weighting` enum. Only
                applicable if method = "conformal".
            weighting_params (dict,optional): Parameters for the selected weighting strategy, if applicable. Only applicable
                if method = "conformal".
                Supported keys include:
                - 'optimization' (str, optional): Past nonconformity score weighting optimization method, as defined in the 'WeightingOptimization' enum.
                - 'decay_param' (float, optional): Exponential decay factor used when 'optimization' is set to 'exponential_decay'.
            threshold_function (str): Method for computing the threshold, as defined in the `ThresholdFunction` enum.
                Only applicable if method = "conformal".
            aggregation (str, int, or None, optional):
                Determines how to aggregate the outlier scores across dimensions for the method outlier_score.
                - If a string (e.g., 'mean', 'max', 'min', 'median'), applies the specified aggregation function over the axes provided in `aggregation_axis`.
                - If an integer, selects the outlier scores corresponding to that specific forecast horizon index.
                - If None, returns outlier scores for all dimensions independently without aggregation.
            aggregation_axis (tuple of int, optional) to consider in outlier_score:
                Specifies the axes over which to aggregate when `aggregation` is a string.
        """

        self.window_size = window_size
        self.quantiles = quantiles
        self.false_alarm = (np.min(quantiles) * 2).item()
        self.nonconformity_score = nonconformity_score
        self.method = method
        self.aggregation = aggregation
        self.aggregation_axis = aggregation_axis

        if self.nonconformity_score in [NonconformityScores.ERROR.value]:
            self.false_alarm = np.min(quantiles).item()

        self.critical_size = np.ceil(1 / self.false_alarm).item()

        if self.method not in [PostHocProbabilisticMethod.CONFORMAL.value, PostHocProbabilisticMethod.GAUSSIAN.value]:
            raise ValueError(f"Provided Post Hoc probabilistic method {self.method} is not valid.")

        if self.nonconformity_score not in [
            NonconformityScores.ABSOLUTE_ERROR.value,
            NonconformityScores.ERROR.value,
        ]:
            raise ValueError(f"Provided nonconformity_score {self.nonconformity_score } is not valid.")

        self.model = kwargs.pop("model", None)
        if self.model is None:
            if self.method == PostHocProbabilisticMethod.CONFORMAL.value:
                self.model = WeightedConformalForecasterWrapper(
                    window_size=self.window_size,
                    false_alarm=self.false_alarm,
                    nonconformity_score=self.nonconformity_score,
                    weighting=weighting,
                    weighting_params=weighting_params,
                    threshold_function=threshold_function,
                )
            elif self.method == PostHocProbabilisticMethod.GAUSSIAN.value:
                self.model = PostHocGaussian(window_size=self.window_size, quantiles=self.quantiles)
            else:
                raise ValueError(f"Invalid method provided {self.method}")

        kwargs["processor_class"] = self.__class__.__name__
        super().__init__(**kwargs)

    def to_dict(self) -> Dict[str, Any]:
        """
        Serializes this instance to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this feature extractor instance.
        """
        output = super().to_dict()

        output["model"] = output["model"].to_dict()

        return output

    def to_json_string(self) -> str:
        """
        Serializes this instance to a JSON string.

        Returns:
            `str`: String containing all the attributes that make up this feature_extractor instance in JSON format.
        """
        dictionary = self.to_dict()

        def recursive_check_ndarray(dictionary):
            for key, value in dictionary.items():
                if key == "dtype":
                    # to do: ensure deserializable
                    dictionary[key] = value.__name__
                elif isinstance(value, np.ndarray):
                    dictionary[key] = value.tolist()
                elif isinstance(value, np.int64):
                    dictionary[key] = int(value)
                elif isinstance(value, list):
                    dictionary[key] = [vv.tolist() if isinstance(vv, np.ndarray) else vv for vv in value]
                elif isinstance(value, dict):
                    dictionary[key] = recursive_check_ndarray(value)
            return dictionary

        dictionary = recursive_check_ndarray(dictionary)

        # make sure private name "_processor_class" is correctly
        # saved as "processor_class"
        _processor_class = dictionary.pop("_processor_class", None)
        if _processor_class is not None:
            dictionary["processor_class"] = _processor_class

        return json.dumps(dictionary, indent=2, sort_keys=True) + "\n"

    @classmethod
    def from_dict(cls, feature_extractor_dict: Dict[str, Any], **kwargs) -> "PreTrainedFeatureExtractor":
        """
        Instantiates a type of [`~feature_extraction_utils.FeatureExtractionMixin`] from a Python dictionary of
        parameters.

        Args:
            feature_extractor_dict (`Dict[str, Any]`):
                Dictionary that will be used to instantiate the feature extractor object. Such a dictionary can be
                retrieved from a pretrained checkpoint by leveraging the
                [`~feature_extraction_utils.FeatureExtractionMixin.to_dict`] method.
            kwargs (`Dict[str, Any]`):
                Additional parameters from which to initialize the feature extractor object.

        Returns:
            [`~feature_extraction_utils.FeatureExtractionMixin`]: The feature extractor object instantiated from those
            parameters.
        """
        model = feature_extractor_dict.get("model", None)
        if model is not None:
            method = feature_extractor_dict.get("method", None)
            if method == PostHocProbabilisticMethod.CONFORMAL.value:
                feature_extractor_dict["model"] = WeightedConformalForecasterWrapper.from_dict(model)
            elif method == PostHocProbabilisticMethod.GAUSSIAN.value:
                feature_extractor_dict["model"] = PostHocGaussian.from_dict(model)
            else:
                raise ValueError(f"Unknown method provided: {method}")

        return super().from_dict(feature_extractor_dict, **kwargs)

    def train(self, y_cal_pred: np.ndarray, y_cal_gt: np.ndarray) -> "PostHocProbabilisticProcessor":
        """Fit posthoc probabilistic wrapper.
        Input:
        y_cal_pred model perdictions: nsamples x forecast_horizon x number_features
        y_cal_gt ground truth values: nsamples x forecast_horizon x number_features

        """
        if len(y_cal_pred.shape) != 3:
            raise ValueError("y_cal_pred should have 3 dimensions: nsamples x forecast_horizon x number_features")

        if len(y_cal_gt.shape) != 3:
            raise ValueError("y_cal_gt should have 3 dimensions: nsamples x forecast_horizon x number_features")

        # WMG to do: check that updated window size is used in the fit call
        # (update) WMG: I don't think we need this
        # if self.window_size is None:
        #     window_size = y_cal_gt.shape[0]
        # else:
        #     window_size = self.window_size

        self.model.fit(
            y_cal_gt=y_cal_gt,
            y_cal_pred=y_cal_pred,  # ttm predicted values
            # ttm corresponding gt values
        )
        return self

    def predict(self, y_test_pred: np.ndarray, quantiles: List[float] = []) -> np.ndarray:
        """Predict posthoc probabilistic wrapper.
        Input:
        y_test_pred: nsamples x forecast_horizon x number_features
        Returns:
        y_test_prob_pred: nsamples x forecast_horizon x number_features x len(quantiles)
        """
        if len(quantiles) == 0:
            quantiles = self.quantiles

        assert (
            len(y_test_pred.shape) == 3
        ), " y_test_pred should have 3 dimensions : nsamples x forecast_horizon x number_features"

        y_test_prob_pred = np.zeros([y_test_pred.shape[0], y_test_pred.shape[1], y_test_pred.shape[2], len(quantiles)])
        if self.method == PostHocProbabilisticMethod.CONFORMAL.value:
            ix_q = 0
            for q in quantiles:
                if self.model.nonconformity_score in [
                    NonconformityScores.ABSOLUTE_ERROR.value,
                    NonconformityScores.ERROR.value,
                ]:
                    if q < 0.5:
                        q_pi_error_rate = q * 2
                        output_q = self.model.predict(y_test_pred, false_alarm=q_pi_error_rate)
                        y_test_prob_pred[..., ix_q] = output_q["prediction_interval"]["y_low"]
                    elif q > 0.5:
                        q_pi_error_rate = (1 - q) * 2
                        output_q = self.model.predict(y_test_pred, false_alarm=q_pi_error_rate)
                        y_test_prob_pred[..., ix_q] = output_q["prediction_interval"]["y_high"]
                    else:
                        if self.model.nonconformity_score in [NonconformityScores.ERROR.value]:
                            q_pi_error_rate = 0.5
                            output_q = self.model.predict(y_test_pred, false_alarm=q_pi_error_rate)
                            y_test_prob_pred[..., ix_q] = output_q["prediction_interval"]["y_high"]
                        else:
                            y_test_prob_pred[..., ix_q] = y_test_pred

                ix_q += 1
        if self.method == PostHocProbabilisticMethod.GAUSSIAN.value:
            y_test_prob_pred = self.model.predict(y_test_pred)

        return y_test_prob_pred

    def update(
        self, y_gt: np.ndarray, y_pred: np.ndarray, X: np.ndarray = None, timestamps: np.ndarray = None
    ) -> "PostHocProbabilisticProcessor":
        """
        Update the probabilistic post hoc model

        Args:
            y_gt (np.ndarray): Ground truth values. Shape: (n_samples, forecast_length, num_features).
            y_pred (np.ndarray): Predicted values. Shape: (n_samples,forecast_length, num_features).
            X (np.ndarray, optional): Input covariates for input-dependent methods. Shape: (n_samples, n_features).
            timestamps (np.ndarray, optional): Timestamps associated with each predicted value. Shape: (n_samples,).
        """
        if self.method == PostHocProbabilisticMethod.CONFORMAL.value:
            self.model.update(y_gt=y_gt, y_pred=y_pred, X=X, timestamps=timestamps)
        if self.method == PostHocProbabilisticMethod.GAUSSIAN.value:
            self.model.update(y_gt=y_gt, y_pred=y_pred)

        return self

    def outlier_score(
        self,
        y_gt: np.ndarray,
        y_pred: np.ndarray,
        X: np.ndarray = None,
        timestamps: np.ndarray = None,
        aggregation: Union[str, int] = -1,
        significance: float = 0.01,
        aggregation_axis: Union[int, Tuple[int, ...]] = -1,
        outlier_label: bool = True,
    ) -> np.ndarray:
        """
        PROTOTYPE: METHOD TO GET NORMALIZED OUTLIER CONFORMAL SCORE (P-VALUE) BASED ON FORECASTED PREDICTION ERRORS

        Args:
            y_gt (np.ndarray): Ground truth values. Shape: (n_samples, forecast_length, num_features).
            y_pred (np.ndarray): Predicted values. Shape: (n_samples,forecast_length, num_features).
            X (np.ndarray, optional): Input covariates for input-dependent methods. Shape: (n_samples, n_features).
            timestamps (np.ndarray, optional): Timestamps associated with each predicted value. Shape: (n_samples,).
            aggregation (str, int, or None, optional):
                Determines how to aggregate the outlier scores across dimensions.
                - If a string (e.g., 'mean', 'max', 'min', 'median'), applies the specified aggregation function over the axes provided in `aggregation_axis`.
                - If an integer, selects the outlier scores corresponding to that specific forecast horizon index.
                - If None, returns outlier scores for all dimensions independently without aggregation.
                - If -1 sets default aggregation in self.aggreggation attribute

            aggregation_axis (tuple of int, optional):
                Specifies the axes over which to aggregate when `aggregation` is a string.
        Returns:
            output_array (np.ndarray): The resulting array after aggregation. Its shape depends on the chosen aggregation strategy, but the last dimension is always 2, representing the outlier score (p-value) and the outlier label.
        """
        if self.method == PostHocProbabilisticMethod.CONFORMAL.value:
            if self.model.nonconformity_score in [
                NonconformityScores.ABSOLUTE_ERROR.value,
            ]:
                output = self.model.predict(
                    y_pred=y_pred, y_gt=y_gt, X=X, timestamps=timestamps, false_alarm=significance
                )
                outliers_scores = output["outliers_scores"]

                """
                Aggregation
                """
                if aggregation == -1:
                    aggregation = self.aggregation
                if aggregation_axis == -1:
                    aggregation_axis = self.aggregation_axis

                if aggregation is None:
                    outliers = output["outliers"]

                elif isinstance(aggregation, int):
                    outliers_scores = self.forecast_horizon_aggregation(outliers_scores, aggregation=aggregation)
                    # outliers_scores = outliers_scores[:, aggregation, :]
                    outliers = np.array(np.array(outliers_scores) <= significance).astype("int")

                elif isinstance(aggregation, str):
                    if aggregation_axis is None:
                        raise ValueError("aggregation_axis must be specified when aggregation is a string")

                    if isinstance(aggregation_axis, int):
                        aggregation_axis = (aggregation_axis,)

                    ### If aggregation across forecast horizon (axis = 1) was selected
                    if 1 in aggregation_axis:
                        aggregation_axis = tuple(x for x in aggregation_axis if x != 1)
                        outliers_scores = self.forecast_horizon_aggregation(outliers_scores, aggregation=aggregation)
                        if aggregation_axis:
                            outliers_scores = outliers_scores[:, np.newaxis, :]

                    ### If aggreagtion axis for other dimension != 1 were selected
                    if aggregation_axis:
                        if aggregation == "mean":
                            outliers_scores = np.mean(outliers_scores, axis=aggregation_axis)
                        elif aggregation == "median":
                            outliers_scores = np.median(outliers_scores, axis=aggregation_axis)
                        elif aggregation == "max":
                            outliers_scores = np.max(outliers_scores, axis=aggregation_axis)
                        elif aggregation == "min":
                            outliers_scores = np.min(outliers_scores, axis=aggregation_axis)
                        else:
                            raise ValueError(f"Unsupported aggregation method: {aggregation}")
                    outliers = np.array(np.array(outliers_scores) <= significance).astype("int")
                else:
                    raise TypeError("aggregation must be either an int or a supported aggregation string")

                if outlier_label:
                    return np.stack([outliers_scores, outliers], axis=-1)
                else:
                    return outliers_scores

    def forecast_horizon_aggregation(
        self, outliers_scores: np.ndarray, aggregation: Union[str, int] = "mean"
    ) -> np.ndarray:
        N, H, F = outliers_scores.shape
        # we want to align predictions/forecasts for each timestamp (observation)
        aligned = np.full(
            (N, H, F), np.nan
        )  # nan for padding initial items for which we have less than horizon H predictions.
        for h in range(H):
            # shift each row of forecast horizon h by h steps into the future.
            aligned[h:N, h, :] = outliers_scores[
                : N - h, h, :
            ]  # item i should have a values for [i,j,:] for j \in [0, min(i,H)]

        if isinstance(aggregation, int):  ## choose a particular forecast horizon score
            out = aligned[:, aggregation, :]
            for h in range(aggregation + 1):  # aligned in an upper triangular NaNs matrix
                out[h, :] = aligned[h, h, :]
            return out

        elif isinstance(aggregation, str):
            # aggregation but ignore nans
            if aggregation == "mean":
                return np.nanmean(aligned, axis=1)
            elif aggregation == "median":
                return np.nanmedian(aligned, axis=1)
            elif aggregation == "max":
                return np.nanmax(aligned, axis=1)
            elif aggregation == "min":
                return np.nanmin(aligned, axis=1)
            else:
                raise ValueError(f"Unsupported aggregation method: {aggregation}")
        else:
            raise TypeError("aggregation must be either an int or a supported aggregation string")


def absolute_error(y: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    """
    Compute the absolute error between `y` and `y_pred`. If the inputs are multi-dimensional, the average is computed over the last dimension.

    Args:
        y (np.ndarray): Ground truth values.
        y_pred (np.ndarray): Predicted values.

    Returns:
        np.ndarray: Absolute error. If inputs are multi-dimensional, returns the mean absolute error over the last axis.
    """

    assert y.shape == y_pred.shape, (
        "Shapes of y and y_pred do not match: y.shape, y_pred.shape = " + str(y.shape) + ";" + str(y_pred.shape)
    )
    error = np.abs(y - y_pred)
    if len(error.shape) > 1:
        error = np.mean(error, axis=-1)
    return error


def error(y: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    """
    Compute the difference between `y` and `y_pred` (signed difference y - y_pred). If the inputs are multi-dimensional, the average is computed over the last dimension.

    Args:
        y (np.ndarray): Ground truth values.
        y_pred (np.ndarray): Predicted values.

    Returns:
        np.ndarray: signed error. If inputs are multi-dimensional, returns the mean absolute error over the last axis.
    """
    assert y.shape == y_pred.shape, (
        "Shapes of y and y_pred do not match: y.shape, y_pred.shape = " + str(y.shape) + ";" + str(y_pred.shape)
    )
    error = y - y_pred
    if len(error.shape) > 1:
        error = np.mean(error, axis=-1)
    return error


def nonconformity_score_functions(
    y_gt: np.ndarray,
    y_pred: np.ndarray,
    X: Optional[np.ndarray] = None,
    nonconformity_score: str = NonconformityScores.ABSOLUTE_ERROR.value,
) -> np.ndarray:
    """
    Compute a predictive nonconformity score between ground truth and predictions with optional input covariates.

    Args:
        y_gt (np.ndarray): Ground truth values.
        y_pred (np.ndarray): Predicted values.
        X (np.ndarray, optional): Input covariates. Used by some nonconformity scoring methods.
        nonconformity_score (str, optional): Type of nonconformity score to compute, as defined in the `NonconformityScores` enum.

    Returns:
        np.ndarray: Computed nonconformity scores.
    """
    assert nonconformity_score in [
        s.value for s in NonconformityScores
    ], "Selected nonconformity score is not supported, choose from {}".format(
        str([s.value for s in NonconformityScores])
    )
    if nonconformity_score == NonconformityScores.ABSOLUTE_ERROR.value:
        return absolute_error(y_gt, y_pred)
    if nonconformity_score == NonconformityScores.ERROR.value:
        return error(y_gt, y_pred)


def conformal_set(
    y_pred: np.ndarray,
    score_threshold: Union[float, List[float]],
    nonconformity_score: str = NonconformityScores.ABSOLUTE_ERROR.value,
) -> Dict[str, np.ndarray]:
    """
    Compute the conformal prediction set or interval for a given prediction.

    Args:
        y_pred (np.ndarray): Predicted values.
        score_threshold (float or List[float]): Quantile(s) of the nonconformity scores used to determine the prediction intervals.
        nonconformity_score (str, optional): Type of nonconformity score to apply, as defined in the `NonconformityScores` enum.

    Returns:
        Dict[str, np.ndarray]: A dictionary containing the lower and upper bounds of the prediction interval or set. Keys are typically 'y_low' and 'y_high' for prediction intervals.
    """

    if nonconformity_score == NonconformityScores.ABSOLUTE_ERROR.value:
        return {"y_low": y_pred - score_threshold, "y_high": y_pred + score_threshold}
    if nonconformity_score == NonconformityScores.ERROR.value:
        return {"y_low": score_threshold[0] + y_pred, "y_high": score_threshold[1] + y_pred}


class PostHocGaussian:
    def __init__(self, window_size=None, quantiles=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]):
        self.window_size = window_size
        self.quantiles = quantiles
        self.critical_size = 1
        self.variance = None
        self.errors = None  # stores the errors, needed for the online updates

        """
        PostHoc Probabilistic Gaussian Wrapper.

        Transforms the point forecasts of a multivariate model into probabilistic forecasts under the assumption of independent Gaussian residuals.

        Args:
            window_size (int, optional): Maximum number of past residuals to consider when estimating variance. If None (default), all available past residuals are used.
            quantiles (List[float]): List of target quantiles to compute. Each value must lie in the open interval (0, 1).
        """

    def fit(self, y_cal_gt: np.ndarray, y_cal_pred: np.ndarray):
        """
        Fit the PostHoc Probabilistic Gaussian Wrapper.

        Args:
            y_cal_gt (np.ndarray): Ground truth values used for calibration.  Shape: (n_samples, forecast_horizon, num_features).
            y_cal_pred (np.ndarray): Model predictions corresponding to the ground truth. Shape: (n_samples, forecast_horizon, num_features).
        """

        if self.window_size is None:
            window_size = y_cal_gt.shape[0]
        else:
            window_size = self.window_size
        assert (
            len(y_cal_pred.shape) == 3
        ), " y_cal_pred should have 3 dimensions : nsamples x forecast_horizon x number_features"
        assert (
            len(y_cal_gt.shape) == 3
        ), " y_cal_gt should have 3 dimensions : nsamples x forecast_horizon x number_features"

        self.errors = y_cal_gt[-window_size:] - y_cal_pred[-window_size:]
        self.variance = np.sum(self.errors**2, axis=0) / (len(self.errors) - 1)  # dimension should be

    def update(self, y_gt: np.ndarray, y_pred: np.ndarray):
        """
        Update the PostHoc Probabilistic Gaussian Wrapper.

        Args:
            y_cal_gt (np.ndarray): Ground truth values used for calibration.  Shape: (n_samples, forecast_horizon, num_features).
            y_cal_pred (np.ndarray): Model predictions corresponding to the ground truth. Shape: (n_samples, forecast_horizon, num_features).
        """
        errors = y_gt - y_pred
        self.errors = np.concatenate([self.errors, errors], axis=0)
        self.errors = self.errors[-self.window_size :]
        self.variance = np.sum(self.errors**2, axis=0) / (len(self.errors) - 1)

    def predict(self, y_test_pred: np.ndarray, quantiles=[]):
        """
        Predict using the PostHoc Probabilistic Gaussian Wrapper.

        Args:
            y_test_pred (np.ndarray): Model point predictions. Shape: (n_samples, forecast_horizon, num_features).

        Returns:
            np.ndarray: Quantile estimates for each prediction. Shape: (n_samples, forecast_horizon, num_features, len(quantiles)).
        """

        if len(quantiles) == 0:
            quantiles = self.quantiles
        assert self.variance is not None, "Method needs to be fitted"
        assert (
            len(y_test_pred.shape) == 3
        ), " y_test_pred should have 3 dimensions: nsamples x forecast_horizon x number_features"
        std_devs = np.sqrt(self.variance)[..., None]  # Standard deviation for each distribution
        quantiles = norm.ppf(quantiles) * std_devs  # Compute quantiles
        y_test_prob_pred = y_test_pred[..., np.newaxis] + quantiles[np.newaxis, ...]
        return y_test_prob_pred

    def to_dict(self) -> Dict[str, Any]:
        """
        Serializes this instance to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this object.
        """

        output = {
            "window_size": self.window_size,
            "quantiles": self.quantiles,
            "critical_size": self.critical_size,
            "variance": self.variance,
        }

        return output

    @classmethod
    def from_dict(cls, params: Dict[str, Any], **kwargs) -> "PostHocGaussian":
        """
        Instantiates a type of [`~PostHocGaussian`] from a Python dictionary of
        parameters.

        Args:
            dict (`Dict[str, Any]`):
                Dictionary that will be used to instantiate the object.
            kwargs (`Dict[str, Any]`):
                Additional parameters from which to initialize the object.

        Returns:
            [`~PostHocGaussian`]: The PostHocGaussian object instantiated from those
            parameters.
        """

        variance = params.pop("variance", None)
        critical_size = params.pop("critical_size", 1)
        obj = cls(**params)
        obj.variance = variance
        obj.critical_size = critical_size
        return obj


class WeightedConformalForecasterWrapper:
    def __init__(
        self,
        nonconformity_score: str = NonconformityScores.ABSOLUTE_ERROR.value,
        false_alarm: float = 0.05,
        weighting: str = Weighting.UNIFORM.value,
        weighting_params: Dict[str, Any] = {},
        threshold_function: str = ThresholdFunction.WEIGHTING.value,
        window_size: Optional[int] = None,
    ):
        self.nonconformity_score = nonconformity_score
        # self.nonconformity_score_func = nonconformity_score_functions

        self.quantile = 1 - false_alarm
        self.false_alarm = false_alarm

        self.weighting = weighting
        self.weighting_params = weighting_params
        self.window_size = window_size
        self.online_size = 1
        self.threshold_function = threshold_function
        self.weights_adaptive = []
        self.univariate_wrappers = {}

        """
        Weighted Split Conformal Forecasting Wrapper Class.

        Args:
            nonconformity_score (str): Type of nonconformity score to use, as defined in the `NonconformityScores` enum.
            false_alarm (float): Desired false alarm (error) rate for the prediction intervals.
            weighting (str): Strategy for weighting nonconformity scores, as defined in the `Weighting` enum.
            weighting_params (dict): Parameters for the selected weighting strategy, if applicable.
            threshold_function (str): Method for computing the threshold, as defined in the `ThresholdFunction` enum.
            window_size (int, optional): Maximum number of past nonconformity scores to use for calibration. Default is None (use all available scores).
        """

    def to_dict(self) -> Dict[str, Any]:
        """
        Serializes this instance to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this object.
        """

        output = {
            "nonconformity_score": self.nonconformity_score,
            "false_alarm": self.false_alarm,
            "weighting": self.weighting,
            "weighting_params": self.weighting_params,
            "window_size": self.window_size,
            "threshold_function": self.threshold_function,
            "weights_adaptive": self.weights_adaptive,
            "univariate_wrappers": {json.dumps(k): v.to_dict() for k, v in self.univariate_wrappers.items()},
        }

        return output

    @classmethod
    def from_dict(cls, params: Dict[str, Any], **kwargs) -> "WeightedConformalForecasterWrapper":
        """
        Instantiates a type of [`~WeightedConformalForecasterWrapper`] from a Python dictionary of
        parameters.

        Args:
            dict (`Dict[str, Any]`):
                Dictionary that will be used to instantiate the object.
            kwargs (`Dict[str, Any]`):
                Additional parameters from which to initialize the object.

        Returns:
            [`~WeightedConformalForecasterWrapper`]: The WeightedcCnformalForecaster object instantiated from those
            parameters.
        """

        params_copy = params.copy()

        save_attrs = ["weights_adaptive", "univariate_wrappers"]
        copy_attrs = ["weights_adaptive"]

        save_attrs_dict = {}
        for attr in save_attrs:
            save_attrs_dict[attr] = params_copy.pop(attr)

        obj = cls(**params_copy, **kwargs)

        for attr_name, attr_value in save_attrs_dict.items():
            if attr_name in copy_attrs:
                setattr(obj, attr_name, attr_value)

        univariate_wrappers = {}
        for k, v in save_attrs_dict["univariate_wrappers"].items():
            k_decoded = tuple(json.loads(k))
            univariate_wrappers[k_decoded] = WeightedConformalWrapper.from_dict(v)

        obj.univariate_wrappers = univariate_wrappers

        return obj

    def fit(
        self, y_cal_gt: np.ndarray, y_cal_pred: np.ndarray, X_cal: np.ndarray = None, cal_timestamps: np.ndarray = None
    ):
        """
        Fit the Weighted Split Conformal Forecasting method.

        Args:
            y_cal_gt (np.ndarray): Ground truth values. Shape: (num_samples, forecast_length, num_features).
            y_cal_pred (np.ndarray): Forecasted values from the time series model. Shape: (num_samples, forecast_length, num_features).
            X_cal (np.ndarray, optional): Input covariates for input-dependent conformal methods. Shape: (num_samples, num_covariates).
            cal_timestamps (np.ndarray, optional): Timestamps associated with each forecasted value. Shape: (num_samples, forecast_length).
        """

        if self.window_size is None:
            self.window_size = y_cal_pred.shape[0]

        window_critical_size = int(np.ceil(1 / self.false_alarm))
        if (self.window_size < window_critical_size) and (y_cal_pred.shape[0] >= window_critical_size):
            self.window_size = window_critical_size
        assert (
            self.window_size >= window_critical_size
        ), "Not enough calibration points for the desired error rate. For an error rate of {} we need at least {} calibration points".format(
            self.false_alarm, window_critical_size
        )

        self.univariate_wrappers = {}
        for ix_f in range(y_cal_pred.shape[2]):
            cal_weights = None

            """
            Weights Optimization
            """
            # Copy the dictionary
            weighting_params = self.weighting_params.copy()
            if "optimization" in self.weighting_params.keys():
                if self.weighting_params["optimization"] == WeightingOptimization.WASS1.value:  # "wass1":
                    weighting_params.pop("optimization", None)
                    """
                    Fitting Weights
                    1. compute non-conformity scores.3
                    2. Initialize Wass1 weight adapter
                    """
                    critical_efficient_size = int(np.ceil(1 / self.false_alarm))
                    cal_scores = nonconformity_score_functions(
                        y_cal_gt[:, :, ix_f],
                        y_cal_pred[:, :, ix_f],
                        X=X_cal,
                        nonconformity_score=self.nonconformity_score,
                    )

                    stride = 1
                    n_batch_update = int((critical_efficient_size + 2))
                    n_updates = 100
                    n_cal_optimization = (n_updates - 1) * stride + n_batch_update
                    n_cal_init = np.maximum(int(critical_efficient_size) * 2, cal_scores.shape[0] - n_cal_optimization)
                    lr = 0.001

                    if cal_scores.shape[0] >= n_cal_init + n_batch_update:
                        awcsw = AdaptiveWeightedConformalScoreWrapper(
                            false_alarm=self.false_alarm,
                            window_size=self.window_size,
                            weighting=Weighting.UNIFORM.value,
                            weighting_params={
                                "n_batch_update": n_batch_update,
                                "conformal_weights_update": False,
                                "stride": stride,
                                "lr": lr,
                            },
                        )

                        # Need to call these because of side effects
                        _ = awcsw.fit(cal_scores[0 : int(n_cal_init)])  # betas_fit
                        _ = awcsw.predict(cal_scores[int(n_cal_init) :])  # betas_update
                        cal_weights = awcsw.cal_weights

            for ix_h in range(y_cal_pred.shape[1]):
                cal_timestamps_i = None
                if cal_timestamps is not None:
                    cal_timestamps_i = cal_timestamps[:, ix_h, ix_f]
                self.univariate_wrappers[ix_h, ix_f] = WeightedConformalWrapper(
                    nonconformity_score=self.nonconformity_score,
                    false_alarm=self.false_alarm,
                    weighting=self.weighting,
                    weighting_params=weighting_params,
                    threshold_function=self.threshold_function,
                    window_size=self.window_size,
                    online_size=self.online_size,
                )
                self.univariate_wrappers[ix_h, ix_f].fit(
                    y_cal_gt[:, ix_h, ix_f], y_cal_pred[:, ix_h, ix_f], X_cal=X_cal, cal_timestamps=cal_timestamps_i
                )
                if cal_weights is not None:
                    self.univariate_wrappers[ix_h, ix_f].weights.append(cal_weights)

    def update(self, y_gt: np.ndarray, y_pred: np.ndarray, X: np.ndarray = None, timestamps: np.ndarray = None):
        """
        Update the nonconformity scores and threshold function.

        Args:
            y_gt (np.ndarray): Ground truth values. Shape: (n_samples, forecast_length, num_features).
            y_pred (np.ndarray): Predicted values. Shape: (n_samples,forecast_length, num_features).
            X (np.ndarray, optional): Input covariates for input-dependent methods. Shape: (n_samples, n_features).
            timestamps (np.ndarray, optional): Timestamps associated with each predicted value. Shape: (n_samples,).
        """
        # print(y_pred.shape, y_gt.shape)
        for ix_h in range(y_pred.shape[1]):
            for ix_f in range(y_pred.shape[2]):
                timestamps_i = None
                if timestamps is not None:
                    timestamps_i = timestamps[:, ix_h, ix_f]

                y_gt_i = y_gt[:, ix_h, ix_f]

                _ = self.univariate_wrappers[ix_h, ix_f].predict(
                    y_pred[:, ix_h, ix_f], y_gt=y_gt_i, X=X, timestamps=timestamps_i, update=True
                )

    def predict(
        self,
        y_pred: np.ndarray,
        y_gt: np.ndarray = None,
        X: np.ndarray = None,
        timestamps: np.ndarray = None,
        false_alarm: float = None,
        update: bool = False,
    ) -> Dict[str, np.ndarray]:
        """
        Generate prediction intervals and, if ground truth (`y_gt`) is provided, optionally return outlier flags and nonconformity p-values.

        If `update` is enabled, the method calls `predict_batch` with the `update=True` flag every `self.online_size` samples to update the nonconformity scores.

        Args:
            y_pred (np.ndarray): Predicted values. Shape: (n_samples,forecast_length, num_features).
            y_gt (np.ndarray, optional): Ground truth values. Shape: (n_samples, forecast_length, num_features).
            X (np.ndarray, optional): Input covariates. Shape: (n_samples, n_features).
            timestamps (np.ndarray, optional): Timestamps associated with each predicted value. Shape: (num_samples, forecast_length ).
            false_alarm (float, optional): Desired error rate in [0, 1]. If None, defaults to `self.false_alarm`.
            update (bool, optional): Whether to update the nonconformity scores using `y_gt` if provided. Default is False.

        Returns:
            dict: A dictionary containing:
                - prediction intervals for each forecast step and feature (always).
                - and optionally, outlier flags and nonconformity p-values if `y_gt` is provided.
        """

        output = {}
        if self.nonconformity_score in [s.value for s in NonconformityScores]:  #'absolute_error':
            output["prediction_interval"] = {"y_low": np.zeros_like(y_pred), "y_high": np.zeros_like(y_pred)}

        if y_gt is not None:
            output["outliers"] = np.zeros_like(y_pred)
            output["outliers_scores"] = np.zeros_like(y_pred)

        for ix_h in range(y_pred.shape[1]):
            for ix_f in range(y_pred.shape[2]):
                timestamps_i = None
                if timestamps is not None:
                    timestamps_i = timestamps[:, ix_h, ix_f]

                y_gt_i = None
                if y_gt is not None:
                    y_gt_i = y_gt[:, ix_h, ix_f]

                output_i = self.univariate_wrappers[ix_h, ix_f].predict(
                    y_pred[:, ix_h, ix_f],
                    y_gt=y_gt_i,
                    X=X,
                    timestamps=timestamps_i,
                    false_alarm=false_alarm,
                    update=update,
                )

                if self.nonconformity_score in [s.value for s in NonconformityScores]:  # 'absolute_error':
                    output["prediction_interval"]["y_low"][:, ix_h, ix_f] = output_i["prediction_interval"]["y_low"]
                    output["prediction_interval"]["y_high"][:, ix_h, ix_f] = output_i["prediction_interval"]["y_high"]

                if y_gt is not None:
                    output["outliers"][:, ix_h, ix_f] = output_i["outliers"]
                    output["outliers_scores"][:, ix_h, ix_f] = output_i["outliers_scores"]

        return output


class WeightedConformalWrapper:
    def __init__(
        self,
        nonconformity_score: str = NonconformityScores.ABSOLUTE_ERROR.value,
        false_alarm: float = 0.05,
        weighting: str = Weighting.UNIFORM.value,
        weighting_params: Dict[str, Any] = {},
        threshold_function: str = ThresholdFunction.WEIGHTING.value,
        window_size: Optional[int] = None,
        online_adaptive: bool = False,
        online_size: Optional[int] = 1,
    ):
        """
        Weighted Split Conformal Univariate Wrapper Class.

        Args:
            nonconformity_score (str): Type of nonconformity score to use, as defined in the `NonconformityScores` enum.
            false_alarm (float): Desired false alarm (error) rate for the prediction intervals.
            weighting (str): Weighting strategy for the nonconformity scores, as defined in the `Weighting` enum.
            weighting_params (dict): Dictionary of parameters for the weighting method, if applicable.
            threshold_function (str): Method used to compute the threshold, as defined in the `ThresholdFunction` enum.
            window_size (int, optional): Maximum number of past nonconformity scores to consider for calibration.
            online_adaptive (bool): Whether the method uses an online/adaptive update strategy.
            online_size (int, optional): Stride or update frequency for online adaptation.
        """
        self.nonconformity_score = nonconformity_score
        assert self.nonconformity_score in [
            s.value for s in NonconformityScores
        ], "Selected nonconformity score is not supported, choose from {}".format(
            str([s.value for s in NonconformityScores])
        )
        # self.nonconformity_score_func = nonconformity_score_functions

        self.quantile = 1 - false_alarm
        self.false_alarm = false_alarm

        self.weighting = weighting
        self.weighting_params = weighting_params
        self.window_size = window_size
        self.online_size = online_size
        self.online = online_adaptive

        self.threshold_function = threshold_function

        self.cal_scores = []
        self.weights = []
        self.cal_X = []
        self.cal_timestamps = []

        self.score_threshold = None

    def to_dict(self) -> Dict[str, Any]:
        """
        Serializes this instance to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this object.
        """
        output = {
            "nonconformity_score": self.nonconformity_score,
            "false_alarm": self.false_alarm,
            "weighting": self.weighting,
            "weighting_params": self.weighting_params,
            "window_size": self.window_size,
            "online_adaptive": self.online,
            "online_size": self.online_size,
            "threshold_function": self.threshold_function,
            "cal_scores": self.cal_scores,
            "weights": self.weights,
            "cal_X": self.cal_X,
            "cal_timestamps": self.cal_timestamps,
            "score_threshold": self.score_threshold,
        }

        return output

    @classmethod
    def from_dict(cls, params: Dict[str, Any], **kwargs) -> "WeightedConformalWrapper":
        """
        Instantiates a type of [`~WeightedConformalWrapper`] from a Python dictionary of
        parameters.

        Args:
            dict (`Dict[str, Any]`):
                Dictionary that will be used to instantiate the object.
            kwargs (`Dict[str, Any]`):
                Additional parameters from which to initialize the object.

        Returns:
            [`~WeightedConformalWrapper`]: The WeightedcCnformalWrapper object instantiated from those
            parameters.
        """

        params_copy = params.copy()

        save_attrs = ["cal_scores", "weights", "cal_X", "cal_timestamps", "score_threshold"]
        save_attrs_dict = {}
        for attr in save_attrs:
            save_attrs_dict[attr] = params_copy.pop(attr)

        obj = cls(**params_copy, **kwargs)

        for attr_name, attr_value in save_attrs_dict.items():
            if isinstance(attr_value, list):
                setattr(obj, attr_name, np.asarray(attr_value))
            else:
                setattr(obj, attr_name, attr_value)

        return obj

    def fit(
        self,
        y_cal_gt: np.ndarray,
        y_cal_pred: np.ndarray,
        X_cal: Optional[np.ndarray] = None,
        cal_timestamps: Optional[np.ndarray] = None,
    ):
        """
        Fit the Weighted Split Conformal method.

        Args:
            y_cal_gt (np.ndarray): Ground truth values. Shape: (num_samples).
            y_cal_pred (np.ndarray): Forecasted values from the time series model. Shape: (num_samples).
            X_cal (np.ndarray, optional): Input covariates used for input-dependent conformal methods. Shape: (num_samples, num_covariates).
            cal_timestamps (np.ndarray, optional): Timestamps corresponding to each forecasted value. Shape: (num_samples).
        """

        if self.window_size is None:
            self.window_size = y_cal_pred.shape[0]
        self.cal_scores = nonconformity_score_functions(
            y_cal_gt, y_cal_pred, X=X_cal, nonconformity_score=self.nonconformity_score
        )

        self.cal_scores = self.cal_scores[-self.window_size :]

        if X_cal is not None:
            self.cal_X = X_cal[-self.window_size :]

        if cal_timestamps is not None:
            self.cal_timestamps = cal_timestamps[-self.window_size :]

        critical_efficient_size = int(np.ceil(1 / self.false_alarm))

        # Certain Weighting Methods May Require Fitting
        if self.weighting in [
            Weighting.UNIFORM.value,
            Weighting.EXPONENTIAL_DECAY.value,
        ]:  # ["uniform", "exponential_decay"]:
            cal_weights = self.get_weights()
            self.weights.append(cal_weights)
            # self.weights.append(cal_weights[-self.cal_scores.shape[0]:])
            if self.threshold_function == ThresholdFunction.WEIGHTING.value:  #  "weighting":
                self.score_threshold = self.score_threshold_func(cal_weights, false_alarm=self.false_alarm)
        assert (
            np.sum(cal_weights) >= critical_efficient_size
        ), " The effective size is too small for the desired false alarm of {}, the calibration set should be larger than {} ".format(
            self.false_alarm, critical_efficient_size
        )

    def get_weights(
        self,
        y_pred: Optional[np.ndarray] = None,
        X: Optional[np.ndarray] = None,
        timestamps: Optional[np.ndarray] = None,
        false_alarm: Optional[float] = None,
    ) -> np.ndarray:
        """
        Return the nonconformity calibration weights.

        Args:
            y_pred (np.ndarray, optional): Predicted values. Shape: (num_samples,).
            X (np.ndarray, optional): Input covariates for input-dependent conformal methods. Shape: (num_samples, num_covariates).
            timestamps (np.ndarray, optional): Timestamps associated with each predicted value. Shape: (num_samples,).
            false_alarm (float, optional): Desired error rate (false alarm rate) for the prediction interval.

        Returns:
            np.ndarray: Array of calibration weights. Shape is either (window_size,) or (num_samples, window_size), depending on the weighting strategy.
        """

        if false_alarm is None:
            false_alarm = self.false_alarm

        if self.weighting in [
            Weighting.UNIFORM.value,
            Weighting.EXPONENTIAL_DECAY.value,
        ]:  # ["uniform", "exponential_decay"]:
            if len(self.weights) > 0:
                return self.weights[-1]
            else:
                if self.weighting == Weighting.UNIFORM.value:
                    # return np.ones(y_pred.shape[0],self.cal_scores.shape[0])
                    # return np.ones(self.cal_scores.shape[0])
                    return np.ones(self.window_size)

                if self.weighting == Weighting.EXPONENTIAL_DECAY.value:
                    decay_param = self.weighting_params.get("decay_param", 0.99)
                    return decay_param ** (self.window_size - np.arange(self.window_size))

    def score_threshold_func(
        self,
        cal_weights,
        cal_scores: Optional[np.ndarray] = None,
        y_pred: Optional[np.ndarray] = None,
        X: Optional[np.ndarray] = None,
        timestamps: Optional[np.ndarray] = None,
        false_alarm: Optional[float] = None,
    ) -> Union[float, List[float]]:
        """
        Compute the nonconformity score threshold corresponding to the desired false alarm (error) rate.

        Args:
            cal_weights (np.ndarray): Calibration weights. Can be either:
                - 1D array of shape (num_calibration_scores,)
                - 2D array of shape (num_samples, num_calibration_scores)
            cal_scores (np.ndarray, optional): Array of nonconformity scores used for threshold computation. If not provided, `self.cal_scores` is used.
            y_pred (np.ndarray, optional): Predicted values. Shape: (num_samples,).
            X (np.ndarray, optional): Input covariates for input-dependent thresholding. Shape: (num_samples, num_covariates).
            timestamps (np.ndarray, optional): Timestamps associated with each predicted value. Shape: (num_samples,).
            false_alarm (float, optional): Desired false alarm (error) rate for the prediction interval.

        Returns:
            Union[float, List[float]]: Threshold value(s) for the nonconformity scores, used to determine the prediction intervals.
        """

        if cal_scores is None:
            cal_scores = self.cal_scores

        if false_alarm is None:
            false_alarm = self.false_alarm

        score_threshold = []
        n_cal_scores = cal_scores.shape[0]
        assert n_cal_scores >= np.ceil(1 / false_alarm), " not enough calibration scores for error rate " + str(
            false_alarm
        )
        if self.threshold_function == ThresholdFunction.WEIGHTING.value:  # "weighting":
            if self.nonconformity_score in [s.value for s in PositiveNonconformityScores]:
                if len(cal_weights.shape) == 1:  # same weights for all y
                    score_threshold = weighted_conformal_quantile(
                        np.append(cal_scores, np.array([np.inf]), axis=0),
                        np.append(cal_weights[-n_cal_scores:], np.array([1]), axis=0),
                        alpha=false_alarm,
                    )
                else:
                    for i in range(cal_weights.shape[0]):
                        score_threshold_i = weighted_conformal_quantile(
                            np.append(cal_scores, np.array([np.inf]), axis=0),
                            np.append(cal_weights[i, -n_cal_scores:], np.array([1]), axis=0),
                            alpha=false_alarm,
                        )
                        score_threshold.append(score_threshold_i)
                    score_threshold = np.array(score_threshold)

            elif self.nonconformity_score == NonconformityScores.ERROR.value:
                if len(cal_weights.shape) == 1:  # same weights for all y
                    cal_scores_infty = np.append(cal_scores, np.array([np.inf, -np.inf]), axis=0)
                    # cal_scores_infty = np.append(np.array([]),cal_scores_infty,axis=0)

                    cal_weights_infty = np.append(cal_weights[-n_cal_scores:], np.array([1, 1]), axis=0)
                    # cal_weights_infty = np.append(np.array([1]),cal_weights_infty,axis=0)

                    score_threshold_low = weighted_conformal_quantile(
                        cal_scores_infty, cal_weights_infty, alpha=np.maximum(false_alarm / 2, 1 - (false_alarm / 2))
                    )

                    score_threshold_up = weighted_conformal_quantile(
                        cal_scores_infty, cal_weights_infty, alpha=np.minimum(false_alarm / 2, 1 - (false_alarm / 2))
                    )
                    assert (
                        score_threshold_up >= score_threshold_low
                    ), " score_threshold_up is not greater than score_threshold_low"

                    score_threshold = [score_threshold_low, score_threshold_up]

                else:
                    for i in range(cal_weights.shape[0]):
                        cal_scores_infty = np.append(cal_scores, np.array([np.inf, -np.inf]), axis=0)
                        # cal_scores_infty = np.append(np.array([-np.inf]),cal_scores_infty,axis=0)

                        cal_weights_infty = np.append(cal_weights[i, -n_cal_scores:], np.array([1, 1]), axis=0)
                        # cal_weights_infty = np.append(np.array([1]),cal_weights_infty,axis=0)

                        # score_threshold_i = weighted_conformal_quantile(np.append(cal_scores,np.array([np.inf]),axis=0),
                        #                     np.append(cal_weights[i,-n_cal_scores:],np.array([1]),axis=0),
                        #                     alpha=false_alarm)

                        score_threshold_low_i = weighted_conformal_quantile(
                            cal_scores_infty,
                            cal_weights_infty,
                            alpha=np.maximum(false_alarm / 2, 1 - (false_alarm / 2)),
                        )

                        score_threshold_up_i = weighted_conformal_quantile(
                            cal_scores_infty,
                            cal_weights_infty,
                            alpha=np.minimum(false_alarm / 2, 1 - (false_alarm / 2)),
                        )
                        score_threshold_i = [score_threshold_low_i, score_threshold_up_i]
                        assert (
                            score_threshold_up_i >= score_threshold_low_i
                        ), " score_threshold_up is not greater than score_threshold_low"

                        score_threshold.append(score_threshold_i)
                    score_threshold = np.array(score_threshold).transpose()  # first dimension is low/up score

            return score_threshold

    def predict_batch(
        self,
        y_pred: np.ndarray,
        y_gt: Optional[np.ndarray] = None,
        X: Optional[np.ndarray] = None,
        timestamps: Optional[np.ndarray] = None,
        false_alarm: Optional[float] = None,
        update: Optional[bool] = None,
    ) -> Dict[str, np.ndarray]:
        """
        Generate prediction intervals for a batch of data and, if ground truth (`y_gt`) is provided, optionally return outliers and nonconformity p-values.

        Args:
            y_pred (np.ndarray): Predicted values. Shape: (n_samples, 1).
            y_gt (np.ndarray, optional): Ground truth values. Shape: (n_samples, 1).
            X (np.ndarray, optional): Input covariates. Shape: (n_samples, n_features).
            false_alarm (float, optional): Desired error rate in [0, 1]. If None, defaults to `self.false_alarm`.
            update (bool, optional): Whether to update the nonconformity scores using `y_gt` if provided. Default is False.

        Returns:
            dict: A dictionary containing:
                - prediction intervals (always),
                - and optionally, outlier flags and nonconformity p-values if `y_gt` is provided.
        """

        if false_alarm is None:
            false_alarm = self.false_alarm

        if update is None:
            update = self.online

        # Weight Computation
        cal_weights = self.get_weights(y_pred, X=X, timestamps=timestamps, false_alarm=false_alarm)

        # Score Threshold
        if (
            (false_alarm == self.false_alarm)
            and (self.weighting in [Weighting.UNIFORM.value])
            and (self.threshold_function in [ThresholdFunction.WEIGHTING.value])
        ):
            score_threshold = self.score_threshold
        else:
            score_threshold = self.score_threshold_func(
                cal_weights, y_pred=y_pred, X=X, timestamps=timestamps, false_alarm=false_alarm
            )

        # Prediction Interval
        prediction_interval = conformal_set(y_pred, score_threshold, nonconformity_score=self.nonconformity_score)

        output = {}
        if y_gt is not None:
            # Compute nonconformity scores of input
            test_scores = nonconformity_score_functions(
                y_gt, y_pred, X=X, nonconformity_score=self.nonconformity_score
            )

            # Outlier Flag and Outlier Scores
            test_outliers = np.array(test_scores > score_threshold).astype("int")
            test_ad_scores = []
            for score in test_scores:
                ad_score = weighted_conformal_alpha(
                    np.append(self.cal_scores, np.array([np.inf]), axis=0),
                    np.append(cal_weights, np.array([1]), axis=0),
                    score,
                )
                # test_ad_scores.append(1 - ad_score)
                test_ad_scores.append(ad_score)  # p-value (significance)

            # Update
            if update:
                self.update(test_scores, X=X, timestamps=timestamps)
            output["outliers"] = test_outliers
            output["outliers_scores"] = np.array(test_ad_scores).flatten()

        output["prediction_interval"] = prediction_interval
        return output

    def predict(
        self,
        y_pred,
        y_gt: Optional[np.ndarray] = None,
        X: Optional[np.ndarray] = None,
        timestamps: Optional[np.ndarray] = None,
        false_alarm: Optional[float] = None,
        update: Optional[bool] = None,
    ) -> Dict[str, np.ndarray]:
        """
        Generate prediction intervals and, if ground truth (`y_gt`) is provided, optionally return outlier flags and nonconformity p-values.

        If `update` is enabled, the method calls `predict_batch` with the `update=True` flag every `self.online_size` samples to update the nonconformity scores.

        Args:
            y_pred (np.ndarray): Predicted values. Shape: (n_samples, 1).
            y_gt (np.ndarray, optional): Ground truth values. Shape: (n_samples, 1).
            X (np.ndarray, optional): Input covariates. Shape: (n_samples, n_features).
            false_alarm (float, optional): Desired error rate in [0, 1]. If None, defaults to `self.false_alarm`.
            timestamps (np.ndarray, optional): Timestamps associated with each predicted value. Shape: (num_samples,).
            update (bool, optional): Whether to update the nonconformity scores using `y_gt` if provided. Default is False.

        Returns:
            dict: A dictionary containing:
                - prediction intervals (always),
                - and optionally, outlier flags and nonconformity p-values if `y_gt` is provided.
        """

        if false_alarm is None:
            false_alarm = self.false_alarm

        if update is None:
            update = self.online

        n_samples = y_pred.shape[0]
        n_batches = int(np.ceil(n_samples / self.online_size))

        if (y_gt is not None) and update:
            """
            Batch Inference
            """
            output = None
            for ix_b in range(n_batches):
                ix_ini = int(ix_b * self.online_size)
                ix_end = np.minimum(int(ix_b * self.online_size + self.online_size), y_pred.shape[0])

                y_pred_b = y_pred[ix_ini:ix_end]
                y_gt_b = y_gt[ix_ini:ix_end]
                X_b = None
                if X is not None:
                    X_b = X[ix_ini:ix_end]
                timestamps_b = None
                if timestamps is not None:
                    timestamps_b = timestamps[ix_ini:ix_end]
                output_b = self.predict_batch(
                    y_pred_b, y_gt=y_gt_b, X=X_b, timestamps=timestamps_b, false_alarm=false_alarm, update=update
                )
                if output is None:
                    output = output_b.copy()
                else:
                    for k in output_b.keys():
                        if k == "prediction_interval":
                            for k2 in output_b[k].keys():
                                output[k][k2] = np.append(output[k][k2], np.array(output_b[k][k2]), axis=0)
                        else:
                            output[k] = np.append(output[k], np.array(output_b[k]), axis=0)
        else:
            output = self.predict_batch(
                y_pred, y_gt=y_gt, X=X, timestamps=timestamps, false_alarm=false_alarm, update=update
            )

        return output

    def predict_interval(
        self,
        y_pred: np.ndarray,
        X: Optional[np.ndarray] = None,
        timestamps: Optional[np.ndarray] = None,
        false_alarm: Optional[float] = None,
    ) -> Dict[str, np.ndarray]:
        """
        Generate prediction intervals. (no update)

        Args:
            y_pred (np.ndarray): Predicted values. Shape: (n_samples, 1).
            X (np.ndarray, optional): Input covariates. Shape: (n_samples, n_features).
            timestamps (np.ndarray, optional): Timestamps associated with each predicted value. Shape: (num_samples,).
            false_alarm (float, optional): Desired error rate in [0, 1]. If None, defaults to `self.false_alarm`.

        Returns:
            dict: A dictionary containing prediction intervals
        """

        if false_alarm is None:
            false_alarm = self.false_alarm

        # Weight Computation
        cal_weights = self.get_weights(y_pred, X=X, timestamps=timestamps, false_alarm=false_alarm)

        # Score Threshold
        score_threshold = self.score_threshold_func(
            cal_weights, y_pred=y_pred, X=X, timestamps=timestamps, false_alarm=false_alarm
        )

        # Prediction Interval
        prediction_interval = conformal_set(y_pred, score_threshold, nonconformity_score=self.nonconformity_score)

        return prediction_interval

    def update(self, scores: np.ndarray, X: Optional[np.ndarray] = None, timestamps: Optional[np.ndarray] = None):
        """
        Update the nonconformity scores and threshold function.

        Args:
            scores (np.ndarray): Nonconformity scores. Shape: (n_samples, 1).
            X (np.ndarray, optional): Input covariates for input-dependent methods. Shape: (n_samples, n_features).
            timestamps (np.ndarray, optional): Timestamps associated with each predicted value. Shape: (n_samples,).
        """

        self.cal_scores = np.append(self.cal_scores, scores, axis=0)
        self.cal_scores = self.cal_scores[-self.window_size :]

        if timestamps is not None:
            self.cal_timestamps.extend(timestamps)
            self.cal_timestamps = self.cal_timestamps[-self.window_size :]
        if X is not None:
            self.cal_X = np.append(self.cal_X, X, axis=0)
            self.cal_X = self.cal_X[-self.window_size :]

        if self.weighting == Weighting.UNIFORM.value:  # "uniform":
            cal_weights = self.get_weights()
            if self.threshold_function == ThresholdFunction.WEIGHTING.value:
                self.score_threshold = self.score_threshold_func(cal_weights, false_alarm=self.false_alarm)


def weighted_conformal_quantile(
    scores: np.ndarray,
    weights: np.ndarray,
    alpha: float = 0.05,
    conformal_correction: bool = False,
    max_score: float = np.inf,
) -> float:
    """
    Predict the weighted conformal quantile.

    Args:
        scores (np.ndarray): Nonconformity scores. Shape: (n,).
        weights (np.ndarray): Weights corresponding to each score. Shape: (n,).
        alpha (float): Significance level for the quantile (e.g., 0.05 for a 95% prediction interval).
        conformal_correction (bool): Whether to apply conformal quantile correction.
        max_score (float): Maximum allowable value for the quantile.

    Returns:
        float: Estimated weighted conformal quantile.
    """

    if weights is None:
        weights = np.ones_like(scores)

    assert np.max(weights) <= 1, "Maximum weight needs to be smaller or equal than 1"
    assert np.min(weights) >= 0, "Minimum weight needs to be greater or equal to 0"
    assert weights.shape[0] == scores.shape[0], " Scores shape does not match weights shape"

    """Add infinite score to the score list"""
    if conformal_correction:
        weights = np.append(weights, np.array([1]))
        scores = np.append(scores, np.array([max_score]))

    # Ensure weights sum to 1
    weights = np.array(weights) / np.sum(weights)

    # Sort the residuals and corresponding weights
    sorted_indices = np.argsort(scores)
    sorted_scores = scores[sorted_indices]
    sorted_weights = weights[sorted_indices]

    # Compute cumulative sum of weights
    cumulative_weights = np.cumsum(sorted_weights)

    # Find the smallest residual where the cumulative weight exceeds 1 - alpha
    quantile_index = np.searchsorted(cumulative_weights, 1 - alpha)
    conformal_quantile = sorted_scores[quantile_index]

    return conformal_quantile


def weighted_conformal_alpha(
    scores: np.ndarray,
    weights: np.ndarray,
    score_observed: np.ndarray,
    conformal_correction: bool = False,
    max_score: float = np.inf,
) -> float:
    """
    Compute the weighted conformal p-value.

    Args:
        scores (np.ndarray): Calibration or previously observed nonconformity scores. Shape: (n,).
        score_observed (float): Nonconformity score of the test (new) observation.
        weights (np.ndarray): Weights for each calibration score. Shape: (n,).
        conformal_correction (bool): Whether to apply conformal correction (e.g., additive smoothing).
        max_score (float): Maximum allowable value for scores (used for clipping or normalization).

    Returns:
        float: Weighted conformal p-value.
    """

    if weights is None:
        weights = np.ones_like(score_observed)

    """Add infinite score to the score list"""
    if conformal_correction:
        weights = np.append(weights, np.array([1]))
        scores = np.append(scores, np.array([max_score]))

    weights = np.array(weights) / np.sum(weights)

    # Sort the residuals and corresponding weights
    sorted_indices = np.argsort(scores)
    sorted_scores = scores[sorted_indices]
    sorted_weights = weights[sorted_indices]

    smaller_scores = sorted_scores > score_observed

    return np.sum(sorted_weights[smaller_scores])


"""
Adaptive Weight Class
"""


class AdaptiveWeightedConformalScoreWrapper:
    def __init__(
        self,
        false_alarm: float = 0.05,
        window_size: Optional[int] = None,
        weighting: str = Weighting.UNIFORM.value,
        weighting_params: Dict[str, Any] = {},
    ):  # , threshold_function = 'weighting'):
        """
        Adaptive Weighted Conformal Score Class.

        Args:
            window_size (int, optional): Maximum number of past nonconformity scores to consider for calibration.
            false_alarm (float, optional): Desired false alarm (error) rate for the prediction intervals.
            weighting (str, optional): Weighting strategy for the nonconformity scores, as defined in the `Weighting` enum.
            weighting_params (dict, optional): Dictionary of parameters for the weighting method, if applicable.
                Supported keys include:
                - 'lr' (float, optional): Learning rate for weight optimization.
                - 'n_batch_update' (int, optional): Number of consecutive observations used to compute the empirical CDF for the Wasserstein-1 distance loss.
                - 'stride' (int, optional): Number of new observations between successive optimization updates.
                - 'epochs' (int, optional): Number of optimization steps performed on the same data batch.
                - 'conformal_weights_update' (bool, optional): Whether to include conformal correction when computing the Wasserstein-1 distance loss for optimization. (Recommended: False for most use cases.)
        """

        self.weighting = weighting
        self.weighting_params = weighting_params
        self.window_size = window_size
        self.weights_optimizer = None  # initializaed in fit
        self.losses = []

        self.cal_scores = []
        self.cal_weights = []

        self.false_alarm = false_alarm
        self.weights_critical_norm = np.ceil(1 / self.false_alarm - 1)
        if self.window_size is not None:
            assert (
                self.window_size > self.weights_critical_norm
            ), "Given the false alarm window size must be larger than {}".format(np.ceil(self.weights_critical_norm))

        if "lr" not in self.weighting_params.keys():
            self.weighting_params["lr"] = 0.001
        if "n_batch_update" not in self.weighting_params.keys():
            self.weighting_params["n_batch_update"] = int(self.weights_critical_norm * 4)
        if "stride" not in self.weighting_params.keys():
            self.weighting_params["stride"] = int(1)
        if "epochs" not in self.weighting_params.keys():
            self.weighting_params["epochs"] = int(1)
        if "conformal_weights_update" not in self.weighting_params.keys():
            self.weighting_params["conformal_weights_update"] = False

        self.weights_average = None
        self.weights_average_count = 0

        assert (
            self.weighting_params["n_batch_update"] > self.weights_critical_norm
        ), "Given the false alarm n_batch_update must be larger than {}".format(np.ceil(self.weights_critical_norm))

    def fit(self, scores: np.ndarray) -> np.ndarray:
        """
        Fit the model using calibration nonconformity scores.

        Args:
            scores (np.ndarray): Nonconformity calibration scores. Shape: (n_samples,).
        """

        """
        1. Add Calibration Scores
        """
        assert scores.shape[0] > self.weights_critical_norm, "Provide a minimum of {} scores for fitting".format(
            self.weights_critical_norm
        )
        self.cal_scores = torch.tensor(scores)
        if self.window_size is None:
            self.window_size = scores.shape[0]

        ## Initialize Weights + Optimizer
        w_init = torch.zeros(self.window_size)
        w_init[-scores.shape[0] :] = 1

        self.weights_parameters = torch.nn.Parameter(w_init)

        self.weights_optimizer = torch.optim.Adam([self.weights_parameters], lr=self.weighting_params["lr"])

        """
        2. Compute p-values for the given scores (should be ignored)
        """

        # Expand to nxn by repeating along the row dimension
        torch_scores_past = self.cal_scores.unsqueeze(0).repeat(scores.shape[0], 1)  # Shape: (n, n)

        with torch.no_grad():
            betas = (
                get_beta(
                    self.cal_scores,
                    torch_scores_past,
                    weights=self.weights_parameters[-scores.shape[0] :],
                    conformal_weights=True,
                )
                .clone()
                .detach()
                .numpy()
            )

        return betas

    def predict(self, scores: np.ndarray, verbose: bool = False) -> np.ndarray:
        """
        Perform online prediction of conformal p-values for observed scores.

        This method updates the weighting parameters in an online fashion based on `self.weighting` and `self.weights_parameters`.

        Args:
            scores (np.ndarray): Observed nonconformity scores. Shape: (n_samples,).
            verbose (bool, optional): If True, enables verbose output. Default is False.


        Returns:
            np.ndarray: Array of weighted p-scores (beta_t) for each test score. Shape: (n_samples,).
        """

        n_scores = scores.shape[0]
        n_batch = int(self.weighting_params["n_batch_update"])
        n_epochs = int(self.weighting_params["epochs"])
        stride = int(self.weighting_params["stride"])
        conformal_weights_update = self.weighting_params["conformal_weights_update"]

        n_batch_updates = (n_scores - n_batch) / stride + 1
        n_batch_updates = int(np.ceil(n_batch_updates))

        losses = []
        beta_output = []
        for i in range(n_batch_updates):
            ini_t_i = int(i * stride)
            end_t_i = int(i * stride + n_batch)
            if verbose:
                LOGGER.info("ini/end batch :", ini_t_i, end_t_i)
            scores_i = torch.tensor(scores[ini_t_i:end_t_i])

            """Generate Matrix With Past Scores per batch observation"""
            scores_concat_i = torch.cat((self.cal_scores, scores_i), dim=0)
            scores_matrix = [
                scores_concat_i[i - self.cal_scores.shape[0] : i]
                for i in range(self.cal_scores.shape[0], len(scores_concat_i))
            ]
            scores_matrix = torch.stack(scores_matrix)

            """Get Betas"""
            if self.weights_average is None:
                self.weights_average = self.weights_parameters.clone().detach()
                self.weights_average_count = 1

            with torch.no_grad():
                if (i < n_batch_updates - 1) & (i > 0):
                    ## Only inference for the stride items (betas should rely on weoghts updated with past observations)
                    betas = (
                        get_beta(
                            scores_i[-stride:],
                            scores_matrix[-stride:, :],
                            weights=self.weights_parameters[-scores_matrix.shape[1] :],
                            # weights=self.weights_average[-scores_matrix.shape[1]:]/self.weights_average_count,
                            conformal_weights=True,
                        )
                        .clone()
                        .detach()
                        .numpy()
                    )
                else:
                    betas = (
                        get_beta(
                            scores_i,
                            scores_matrix,
                            weights=self.weights_parameters[-scores_matrix.shape[1] :],
                            # weights=self.weights_average[-scores_matrix.shape[1]:]/self.weights_average_count,
                            conformal_weights=True,
                        )
                        .clone()
                        .detach()
                        .numpy()
                    )
                if verbose:
                    LOGGER.info("Beta shape :", betas.shape)
                beta_output.extend(betas.tolist())

            """Optimize Weights"""
            losses_i = []
            for _ in range(n_epochs):
                self.weights_optimizer.zero_grad()
                loss = get_w1_distance(
                    scores_i,
                    scores_matrix,
                    weights=self.weights_parameters[-scores_matrix.shape[1] :],
                    conformal_weights=conformal_weights_update,
                )
                losses_i.append(loss.item())

                loss.backward()
                self.weights_optimizer.step()

                ## Project weights into the simplex
                self.weights_parameters.data = project_l1_box_torch(
                    self.weights_parameters.data, min_l1_norm=self.weights_critical_norm, max_value=1.0
                )
                if verbose:
                    LOGGER.info(
                        "loss {}, w_norm_l1 {}, w_min {}, w_max {}".format(
                            loss.item(),
                            self.weights_parameters.data.sum(),
                            self.weights_parameters.data.min(),
                            self.weights_parameters.data.max(),
                        )
                    )

                self.weights_average = self.weights_average + self.weights_parameters.clone().detach()
                self.weights_average_count += 1
                # self.weights_parameters.data.clamp_(0, 1.0)
            losses.append(losses_i)

            """ Update Past Calibration Scores """
            self.cal_scores = torch.cat((self.cal_scores, scores_i[0:stride]), dim=0)
            self.cal_scores = self.cal_scores[-self.window_size :]
            # self.cal_weights = self.weights_average.clone().detach().numpy()/self.weights_average_count
            self.cal_weights = self.weights_parameters.clone().detach().numpy()

        self.losses.extend(losses)

        return np.array(beta_output)


"""
Optimization functions
"""


def get_beta(
    test_scores: torch.Tensor,
    observed_scores: torch.Tensor,
    weights: Optional[torch.Tensor] = None,
    conformal_weights: bool = True,
) -> torch.Tensor:
    """
    Compute the weighted conformal p-scores (beta_t) for a batch of test scores given observed calibration scores.

    Args:
        test_scores (np.ndarray): Array of test (observed) nonconformity scores. Shape: (b,).
        observed_scores (np.ndarray): Array of previously observed scores. Shape: (b, w), where each row contains the calibration scores for a given test point.
        weights (np.ndarray): Array of time-based weights applied to observed scores. Shape: (w,). All weights should lie in the range [0, 1].
        conformal_weights (bool, optional): Apply conformal quantile conservative correction.

    Returns:
        np.ndarray: Array of weighted p-scores (beta_t) for each test score. Shape: (b,).
    """

    b, w = observed_scores.shape

    assert (
        test_scores.ndim == 1 and test_scores.shape[0] == b
    ), "incorrect dimensions for test and observed scores, got {} and {}".format(
        test_scores.shape, observed_scores.shape
    )

    if weights is None:
        weights = torch.ones(w)

    assert (
        weights.ndim == 1 and weights.shape[0] == w
    ), "incorrect dimensions for weights and observed scores, got {} and {}".format(
        weights.shape, observed_scores.shape
    )
    assert (
        weights.max() <= 1 and weights.min() >= 0
    ), "weights should be constrained between 0 and 1, got {} and {}".format(weights.max(), weights.min())

    betas = (weights.unsqueeze(0) * (observed_scores <= test_scores.unsqueeze(1))).sum(1)

    weight_norm = weights.sum()
    if conformal_weights:
        weight_norm = weight_norm + 1
        betas = betas + 1
    betas = betas / weight_norm

    # betas = ((weights.unsqueeze(0) * (observed_scores<= test_scores.unsqueeze(1))).sum(1)+1)/weight_norm
    return betas


def w1_distance_from_betas(beta: torch.Tensor) -> torch.Tensor:
    """
    Compute the 1-Wasserstein distance between the empirical distribution
    given by sorted samples `beta` in [0,1] and the uniform distribution on [0,1].

    Args:
        beta: 1D tensor of shape (N,), sorted (beta[0] <= beta[1] <= ... <= beta[N-1])
              with values in [0,1].

    Returns:
        A scalar tensor representing W1 distance.
    """
    # Number of samples
    N = beta.shape[0]

    # For k = 1..N, define the subinterval [a_k, b_k] = [ (k-1)/N, k/N ]
    k = torch.arange(1, N + 1, dtype=beta.dtype, device=beta.device)
    a = (k - 1) / N  # left endpoints
    b = k / N  # right endpoints

    # ----- Piecewise closed-form integrals -----
    #
    # 1) If beta_k < a_k, then integral = ∫_{a_k}^{b_k} (p - beta_k) dp
    left_val = 0.5 * (b**2 - a**2) - beta * (b - a)

    # 2) If beta_k > b_k, then integral = ∫_{a_k}^{b_k} (beta_k - p) dp
    right_val = beta * (b - a) - 0.5 * (b**2 - a**2)

    # 3) If a_k <= beta_k <= b_k, split at beta_k:
    #    ∫ (|beta_k - p| dp) = (area left of beta_k) + (area right of beta_k)
    #    = beta^2 + 0.5(a^2 + b^2) - beta(a + b)
    middle_val = beta**2 + 0.5 * (a**2 + b**2) - beta * (a + b)

    # ----- Select which formula applies for each i -----
    mask_left = beta < a
    mask_right = beta > b

    val = torch.where(mask_left, left_val, torch.where(mask_right, right_val, middle_val))

    # Summation over i=1..N gives W1 distance
    return val.sum()


def get_w1_distance(test_scores, observed_scores, weights=None, conformal_weights=True) -> float:
    """
    Compute the 1-Wasserstein distance (W1) between the empirical distribution of weighted conformal p-values and the uniform distribution on [0, 1].

    The p-values are estimated using `test_scores` and `observed_scores`, weighted by `weights`.

    Args:
        test_scores (np.ndarray): Test nonconformity scores. Shape: (b,).
        observed_scores (np.ndarray): Previously observed calibration scores. Shape: (b, w), where each row corresponds to calibration scores for a test instance.
        weights (np.ndarray): Time-based weights applied to the calibration scores. Shape: (w,). Values must lie in the range [0, 1].
        conformal_weights (bool, optional): Whether to apply a conservative conformal correction when estimating p-values.

    Returns:
        float: Scalar representing the 1-Wasserstein distance between the empirical p-value distribution and the uniform distribution on [0, 1].
    """

    beta = get_beta(test_scores, observed_scores, weights, conformal_weights=conformal_weights)
    # Sort beta to enforce monotonicity
    beta_sorted = torch.sort(beta)[0]
    return w1_distance_from_betas(beta_sorted)


"""
Weight Projections
"""


def euclidean_proj_simplex_torch(vector: torch.Tensor, radius: int = 1) -> torch.Tensor:
    """Euclidean projection on a positive simplex.
    Algorithm 1 in
    Efficient Projections onto the .1-Ball for Learning in High Dimensions, Duchi et al.
    http://www.cs.berkeley.edu/~jduchi/projects/DuchiSiShCh08.pdf
    """
    assert radius > 0, "radius parameter must be strictly positive"
    n = vector.shape[0]
    # Check if vector is on the simplex
    if torch.isclose(vector.sum(), torch.tensor(radius)) and torch.all(vector >= 0):
        return vector
    v_sorted, _ = torch.sort(vector, descending=True)  # Sort vector in decreasing order
    cum_vector = torch.cumsum(v_sorted, dim=0)
    rho = torch.nonzero(
        v_sorted * torch.arange(1, n + 1, dtype=vector.dtype, device=vector.device) > (cum_vector - radius)
    )[-1]  # Find rho
    theta = (cum_vector[rho] - radius) / (rho + 1)  # Compute theta
    w = torch.clamp(vector - theta, min=0)  # Compute projection
    return w


def project_l1_box_torch(v, min_l1_norm: float = 1, max_value: float = 1) -> torch.Tensor:
    w = torch.clamp(v, min=0, max=max_value)
    if v.sum() > torch.tensor(min_l1_norm):
        return w
    else:
        condition = False
        n_iter = 0
        w_i = torch.clamp(v, max=max_value)  # clamp maximum values
        while not condition:
            w_proj = w_i[w_i < max_value]  # consider only entries that are allowed to grow
            l1_reduce = (
                min_l1_norm - torch.sum(w_i == max_value) * max_value
            )  # adjust l1 norm constrain based on norm of entries that are already in their max values
            if l1_reduce <= 0:
                return torch.clamp(w_i, min=0)

            w_proj = euclidean_proj_simplex_torch(w_proj, radius=l1_reduce)  # project
            w_i[w_i < max_value] = torch.clamp(w_proj, max=max_value, min=0)  # assign

            if torch.sum(w_i) >= torch.tensor(min_l1_norm):
                return w_i
            w_i[w_i == 0] = v[w_i == 0]  ## assign original values to items truncated to zero

            n_iter += 1
            condition = n_iter > 10
            LOGGER.info("iter : ", n_iter)

        w_i = torch.clamp(w_i, max=max_value, min=0)
        LOGGER.info(
            "proj ending due to max iterations, norm_l1 is {} and objective was >= {}".format(w_i.sum(), min_l1_norm)
        )
        return w_i



# Copyright contributors to the TSFM project
#
import math
from typing import Literal, Optional, Union

import torch
from torch.utils.data import Dataset

from ..modeling_tspulse import TSPulseForClassification, TSPulseForReconstruction


def patchwise_stitched_reconstruction(
    model,
    past_values,
    patch_size,
    keys_to_stitch,
    keys_to_aggregate,
    reconstruct_start,
    reconstruct_end,
    debug=False,
):
    """
    Performs patchwise reconstruction within a specified time window on a multivariate time-series tensor.
    Only patches whose **start indices fall within** [reconstruct_start, reconstruct_end) are masked and reconstructed.
    Results are then stitched or aggregated across the specified keys.

    Args:
        model: A callable that accepts `past_values` and `past_observed_mask` and returns a dict of outputs.
        past_values (torch.Tensor): Input tensor of shape [B, L, C], where
                                    B = batch size,
                                    L = sequence length,
                                    C = number of input channels.
        patch_size (int): Size of each non-overlapping patch.
        keys_to_stitch (List[str]): Keys in the model output dict that should be reconstructed and stitched
                                    back into full time-series format (shape [B, L, C]).
        keys_to_aggregate (List[str]): Keys in the model output dict that should be aggregated (mean-pooled)
                                       across patches (resulting in shape [B, ...]).
        reconstruct_start (int): Start index (inclusive) of the reconstruction window.
        reconstruct_end (int): End index (exclusive) of the reconstruction window.
        debug (bool): If True, also returns patch-level outputs and patch indices.

    Returns:
        result_dict (dict): Dictionary containing:
            - For each key in `keys_to_stitch`: Reconstructed and stitched tensor of shape [B, L, C],
              with reconstructed values only in the patchwise regions; rest is filled with NaNs.
            - For each key in `keys_to_aggregate`: Aggregated (mean) tensor of shape [B, ...].

        If `debug=True`, also returns:
            patch_outputs (dict):
                - For each stitched key: Tensor of shape [B, num_selected_patches, patch_size, C],
                  which contains patch-level reconstructed outputs.
                - For each aggregated key: Tensor of shape [B, num_selected_patches, ...],
                  representing the output for each patch before aggregation.
                - "patch_starts_selected" (torch.Tensor): 1D tensor of shape [num_selected_patches]
                  with the start positions of each selected patch (in original time indices).

    Raises:
        ValueError: If no patches fall within the reconstruction window.
    """

    B, L, C = past_values.shape
    device = past_values.device
    num_patches = L // patch_size

    patch_indices = torch.arange(num_patches, device=device)
    patch_starts = patch_indices * patch_size
    # patch_ends = patch_starts + patch_size

    # Select only patches fully inside the reconstruct window
    valid_mask = (patch_starts >= reconstruct_start) & (patch_starts < reconstruct_end)

    selected_patch_indices = patch_indices[valid_mask]
    num_selected_patches = selected_patch_indices.shape[0]

    if num_selected_patches == 0:
        raise ValueError("No patches fall entirely within the reconstruction window.")

    # Step 1: Create expanded inputs → shape: [B * num_selected_patches, L, C]
    past_values_expanded = past_values.unsqueeze(1).repeat(1, num_selected_patches, 1, 1)
    past_values_expanded = past_values_expanded.view(B * num_selected_patches, L, C)

    # Step 2: Create past_observed_mask: [B * num_selected_patches, L, C]
    past_observed_mask = torch.ones_like(past_values, dtype=torch.bool)
    past_observed_mask = past_observed_mask.unsqueeze(1).repeat(1, num_selected_patches, 1, 1)
    past_observed_mask = past_observed_mask.view(B * num_selected_patches, L, C)

    patch_starts_selected = patch_starts[valid_mask]
    for i, start in enumerate(patch_starts_selected):
        b_indices = torch.arange(B, device=device)
        idx_range = slice(start.item(), start.item() + patch_size)
        flat_idx = b_indices * num_selected_patches + i
        past_observed_mask[flat_idx, idx_range, :] = 0  # Mask the patch

    # Step 3: Forward pass
    model_outputs = model(
        past_values=past_values_expanded,
        past_observed_mask=past_observed_mask,
    )

    # Step 4: Prepare indices for patch extraction
    indices = torch.arange(patch_size, device=device).unsqueeze(0) + patch_starts_selected.view(-1, 1)
    indices_expanded = indices.unsqueeze(-1).expand(-1, -1, C)  # [num_selected_patches, patch_size, C]
    indices_expanded = indices_expanded.repeat(B, 1, 1)  # [B * num_selected_patches, patch_size, C]

    # Step 5: Initialize result dict
    result_dict = {key: torch.full_like(past_values, float("nan")) for key in keys_to_stitch}
    patch_outputs = {}

    # Step 6: Stitch keys
    for key in keys_to_stitch:
        output = model_outputs[key]  # [B * num_selected_patches, L, C]
        patches = torch.gather(output, dim=1, index=indices_expanded)  # [B * num_selected_patches, patch_size, C]
        patches = patches.view(B, num_selected_patches, patch_size, C)

        for i, start in enumerate(patch_starts_selected):
            result_dict[key][:, start : start + patch_size, :] = patches[:, i]

        if debug:
            patch_outputs[key] = patches  # [B, num_selected_patches, patch_size, C]

    # Step 7: Aggregate keys
    for key in keys_to_aggregate:
        output = model_outputs[key]  # [B * num_selected_patches, ...]
        out_shape = output.shape[1:]
        output = output.view(B, num_selected_patches, *out_shape)
        mean_output = output.mean(dim=1)
        result_dict[key] = mean_output
        if debug:
            patch_outputs[key] = output  # [B, num_selected_patches, ...]

    if debug:
        patch_outputs["patch_starts_selected"] = patch_starts_selected
        return result_dict, patch_outputs
    else:
        return result_dict


class PatchMaskingDatasetWrapper(Dataset):
    def __init__(self, base_dataset, window_length, patch_length, window_position="first"):
        """
        A dataset wrapper for fine-tuning TSPulse on time-series anomaly detection (AD) tasks
        using patch-level masking.

        In AD fine-tuning, TSPulse is trained to reconstruct specific patches of the input sequence,
        given the context of the surrounding time steps. This wrapper enables systematic patch masking
        across a fixed-size window (first or last) of the input sequence, allowing the model to learn local patterns
        and detect deviations during reconstruction.

        Each original time-series sample is expanded into multiple samples, one for each patch in the
        selected window. For each of these, a different patch is masked.
        using a boolean `past_observed_mask`, while the rest of the input remains visible to the model.

        Specifically, for each input sample of shape [T, C], the wrapper:
        - selects a fixed window of length `window_length` from either the start or end of the sequence
        - splits the window into `ceil(window_length / patch_length)` non-overlapping patches
        - generates multiple copies of the sample, each with a different patch marked as unobserved
            in the `past_observed_mask`, while all other time steps remain visible

        Args:
            base_dataset: The original dataset
            window_length: Number of time steps to apply patch masking to
            patch_length: Length of each patch
            window_position: "first" or "last" — which part of the sequence to apply masking on
        """
        assert window_position in [
            "first",
            "last",
        ], "window_position must be 'first' or 'last'"
        self.base_dataset = base_dataset
        self.window_length = window_length
        self.patch_length = patch_length
        self.window_position = window_position
        self.num_patches = math.ceil(window_length / patch_length)
        self.total_len = len(base_dataset) * self.num_patches

    def __len__(self):
        return self.total_len

    def __getitem__(self, idx):
        base_idx = idx // self.num_patches
        patch_idx = idx % self.num_patches

        item = self.base_dataset[base_idx]
        past_values = item["past_values"]
        T, C = past_values.shape

        # Determine window start position based on window_position
        if self.window_position == "first":
            window_start = 0
        else:  # "last"
            window_start = max(0, T - self.window_length)

        # Calculate actual patch position within full time series
        patch_start = window_start + patch_idx * self.patch_length
        patch_end = min(patch_start + self.patch_length, T)

        # Create full-size mask, mask only selected patch in the selected window
        past_observed_mask = torch.ones((T, C), dtype=torch.bool)
        past_observed_mask[patch_start:patch_end] = 0

        return {
            **item,
            "past_observed_mask": past_observed_mask,
        }


def get_embeddings(
    model: Union[TSPulseForReconstruction, TSPulseForClassification],
    past_values: torch.Tensor,
    past_observed_mask: Optional[torch.Tensor] = None,
    component: Literal["backbone", "decoder"] = "decoder",
    mode: Literal["time", "fft", "register", "full"] = "register",
) -> torch.Tensor:
    """
    Obtain embeddings from a TSPulse model.

    Args:
        model: A TSPulse model object with a callable interface that accepts `past_values` and `past_observed_mask`.
        past_values (torch.Tensor): Input tensor of shape [B, L, C], where
                                    B = batch size,
                                    L = sequence length,
                                    C = number of input channels.
        past_observed_mask (Optional[torch.Tensor]): Mask tensor of the same shape as `past_values`, indicating observed (1.0) vs. missing (0.0) values.
        component (str): Component to use to get the embedding. Allowed values are "backbone" and "decoder".
        mode (str): Specifies the type of embeddings to extract. One of:
            - "time": Extracts time-domain embeddings.
            - "fft": Extracts frequency-domain (FFT) embeddings.
            - "register": Extracts register token embeddings.
            - "full": Returns the full embedding without slicing.

    Returns:
        embeddings (torch.Tensor): Tensor of shape [B, C, D], where D depends on the selected mode.
    """

    num_reg_tokens = model.config.patch_register_tokens
    embeddings = model(past_values, past_observed_mask=past_observed_mask)
    if component == "backbone":
        d_model = model.config.d_model_layerwise[-1]
        num_patches = model.config.num_patches_layerwise[-1]

        embeddings = embeddings["backbone_hidden_state"]  # [B, C, D]
    elif component == "decoder":
        d_model = model.config.decoder_d_model_layerwise[-1]
        num_patches = model.config.decoder_num_patches_layerwise[-1]

        embeddings = embeddings["decoder_hidden_state"]  # [B, C, D]
    else:
        raise ValueError(f"Invalid component: {component}. Choose 'backbone' or 'decoder'.")

    time_emb_size = fft_emb_size = (num_patches // 2) * d_model
    reg_emb_size = num_reg_tokens * d_model

    if mode == "time":
        return embeddings[:, :, :time_emb_size]
    elif mode == "fft":
        return embeddings[:, :, time_emb_size : (time_emb_size + fft_emb_size)]
    elif mode == "register":
        return embeddings[:, :, -reg_emb_size:]
    elif mode == "full":
        return embeddings
    else:
        raise ValueError(f"Invalid mode: {mode}. Choose from 'time', 'fft', 'register', 'full'.")



# Copyright contributors to the TSFM project
#

from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import pytest
from transformers import PatchTSTForPrediction

from tsfm_public.models.tinytimemixer import TinyTimeMixerForPrediction
from tsfm_public.toolkit.util import select_by_index

from ..util import nreps


@pytest.fixture(scope="module")
def ts_data():
    df = pd.DataFrame(
        {
            "id": nreps(["A", "B", "C"], 50),
            "id2": nreps(["XX", "YY", "ZZ"], 50),
            "timestamp": [datetime(2021, 1, 1) + timedelta(days=i) for i in range(50)] * 3,
            "value1": range(150),
            "value2": np.arange(150) ** 2 / 3 + 10,
        }
    )
    return df


@pytest.fixture(scope="module")
def sample_data():
    df = pd.DataFrame(
        {
            "val": range(10),
            "val2": [x + 100 for x in range(10)],
            "cat": ["A", "B"] * 5,
            "cat2": ["CC", "DD", "EE", "FF", "GG"] * 2,
        }
    )
    return df


@pytest.fixture(scope="module")
def ts_data_runs():
    df = pd.DataFrame(
        {
            "run_id": nreps(["1", "2", "3", "4"], 50),
            "asset_id": nreps(["foo", "bar", "foo", "bar"], 50),
            "timestamp": [datetime(2021, 1, 1) + timedelta(days=i) for i in range(50)] * 4,
            "value1": range(200),
        }
    )
    return df


@pytest.fixture(scope="module")
def ts_data_nested():
    series_length = 20
    series_set_1 = [pd.Series(np.arange(series_length) / (i + 1)) for i in range(15)]

    series_set_2 = [np.sin(np.pi * s) for s in series_set_1]

    df = pd.DataFrame(
        {
            "time_int": range(15),
            "id": ["A", "B", "C"] * 5,
            "val": series_set_1,
            "val2": series_set_2,
            "label": [s.iloc[-1] for s in series_set_2],
        }
    )

    def label_fun(x):
        if x > 0.75:
            return "blue"
        if x > 0:
            return "red"
        if x < -0.7:
            return "green"
        else:
            return "yellow"

    df.label = df.label.apply(label_fun)
    df["label_encoded"] = pd.Categorical(df.label).codes

    df["time_date"] = df["time_int"] * timedelta(days=1) + datetime(2020, 1, 1)
    return df


@pytest.fixture(scope="package")
def etth_data_base():
    timestamp_column = "date"
    dataset_path = "https://raw.githubusercontent.com/zhouhaoyi/ETDataset/main/ETT-small/ETTh2.csv"
    data = pd.read_csv(
        dataset_path,
        parse_dates=[timestamp_column],
    )
    return data


@pytest.fixture(scope="package")
def etth_data(etth_data_base):
    timestamp_column = "date"
    id_columns = []
    target_columns = ["HUFL", "HULL", "MUFL", "MULL", "LUFL", "LULL", "OT"]
    prediction_length = 96

    data = etth_data_base.copy()
    train_end_index = 12 * 30 * 24

    context_length = 512  # model.config.context_length

    test_end_index = 12 * 30 * 24 + 8 * 30 * 24
    test_start_index = test_end_index - context_length - 4

    train_data = select_by_index(
        data,
        id_columns=id_columns,
        start_index=0,
        end_index=train_end_index,
    )
    test_data = select_by_index(
        data,
        id_columns=id_columns,
        start_index=test_start_index,
        end_index=test_end_index,
    )

    params = {
        "timestamp_column": timestamp_column,
        "id_columns": id_columns,
        "target_columns": target_columns,
        "prediction_length": prediction_length,
        "context_length": context_length,
    }

    return train_data, test_data, params


@pytest.fixture(scope="package")
def patchtst_base_model():
    model_path = "ibm/test-patchtst"
    model = PatchTSTForPrediction.from_pretrained(model_path)

    return model


@pytest.fixture(scope="module")
def ttm_base_model():
    model_path = "ibm/test-ttm-v1"

    return TinyTimeMixerForPrediction.from_pretrained(model_path)


@pytest.fixture(scope="module")
def ttm_model():
    model_path = "ibm/test-ttm-v1"

    def ttm_model_func(**kwargs):
        return TinyTimeMixerForPrediction.from_pretrained(model_path, **kwargs)

    return ttm_model_func



# Copyright contributors to the TSFM project
#
"""Hugging Face Pipeline for Time Series Anomaly Detection"""

import inspect
from collections import defaultdict
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import torch
from torch import nn as nn
from torch.utils.data import DataLoader
from transformers.data.data_collator import default_data_collator
from transformers.modeling_utils import PreTrainedModel
from transformers.pipelines.base import GenericTensor, build_pipeline_init_args
from transformers.trainer_utils import RemoveColumnsCollator
from transformers.utils.doc import add_end_docstrings
from transformers.utils.generic import ModelOutput

from tsfm_public.models.tinytimemixer.modeling_tinytimemixer import TinyTimeMixerForPrediction
from tsfm_public.models.tinytimemixer.utils.ad_helpers import TinyTimeMixerADUtility
from tsfm_public.models.tspulse.modeling_tspulse import TSPulseForReconstruction
from tsfm_public.models.tspulse.utils.ad_helpers import TSPulseADUtility
from tsfm_public.toolkit.conformal import PostHocProbabilisticProcessor

from .ad_helpers import AnomalyScoreMethods
from .dataset import ForecastDFDataset
from .time_series_forecasting_pipeline import TimeSeriesPipeline


class AggregationFunction(Enum):
    """Enum type aggregation functions used when combining anomaly scores."""

    MIN = "min"
    MAX = "max"
    MEAN = "mean"


def score_smoothing(
    x: np.ndarray,
    smoothing_window_size: int,
) -> np.ndarray:
    """Utility function for moving average smoothing of N-dimensional dataset.
    Smoothing is applied along axis=0.

    Args:
        x   (np.ndarray): numpy array of arbitrary dimension
        smoothing_window_size (int): parameter specifies moving window size used for smoothing
    """
    if smoothing_window_size < 2:
        return x
    elif x.ndim == 1:
        return np.convolve(x, np.ones(smoothing_window_size) / smoothing_window_size, mode="same")
    else:
        return np.array([score_smoothing(x[..., i], smoothing_window_size) for i in range(x.shape[-1])]).T


@add_end_docstrings(
    build_pipeline_init_args(has_tokenizer=False, has_feature_extractor=True, has_image_processor=False)
)
class TimeSeriesAnomalyDetectionPipeline(TimeSeriesPipeline):
    """Time Series Anomaly Detection using HF time series models. This pipeline consumes a `pandas.DataFrame`
    containing the time series data and produces a new `pandas.DataFrame` with anomaly scores.
    """

    def __init__(
        self,
        model: PreTrainedModel,
        *args,
        prediction_mode: Optional[List[str]] = None,
        aggr_function: str = AggregationFunction.MAX.value,
        aggregation_length: int = 32,
        smoothing_length: int = 8,
        probabilistic_processor: Optional[PostHocProbabilisticProcessor] = None,
        **kwargs,
    ):
        """Huggingface pipeline for time series anomaly detection using time series foundation models.

        Args:
            model (PreTrainedModel): time series foundation model instance
            prediction_mode (list, optional): specify list of appropriate modes for anomaly scoring. Defaults to [AnomalyPredictionModes.PREDICTIVE.value].
            aggr_function (str, optional): aggregation function for merging scores using different mode, supported values are (max/min/mean). Defaults to "max".
            aggregation_length (int, optional): parameter required for imputation or window based scoring. Defaults to 32.
            smoothing_length (int, optional): window size for post processing of the generated scores. Defaults to 8.
            probabilistic_processor (PostHocProbabilisticProcessor, optional): if prediction mode is "probabilistic", use the probabilistic processor to determine the p-values associated with the forecasts from the underlying model. Defaults to None.
            expand_score (bool, optional): if true report anomaly score for each target column separately. Defaults to False.
            report_mode (bool, optional): if true reports which prediction mode is detects higher anomaly score for each observation. Defaults to False.
            predictive_score_smoothing (bool, optional): if true smoothing is applied to the forecast score. Defaults to False.
            least_significant_scale (float, optional): value between (0, 1). Model scores are function of data variance, this factor specifies a relative score threshold to data variance for marking anomaly. Defaults to 0.01.
            least_significant_score (float, optional): minimum score for marking anomaly. Defaults to 0.2.

        Raises:
            ValueError: unsupported model
            ValueError: invalid prediction_mode
            ValueError: no pytorch support
        """
        if prediction_mode is None:
            prediction_mode = [AnomalyScoreMethods.PREDICTIVE.value]

        if isinstance(prediction_mode, str):
            prediction_mode = [prediction_mode]

        model_processor = None
        if isinstance(model, TSPulseForReconstruction):
            model_processor = TSPulseADUtility(
                model, mode=prediction_mode, aggregation_length=aggregation_length, **kwargs
            )
        elif isinstance(model, TinyTimeMixerForPrediction):
            model_processor = TinyTimeMixerADUtility(
                model=model, mode=prediction_mode, probabilistic_processor=probabilistic_processor, **kwargs
            )
        else:
            raise ValueError(f"Error: does not support {self.model.__class__} object!")

        kwargs["context_length"] = model.config.context_length
        kwargs["prediction_length"] = model.config.prediction_length

        if "batch_size" not in kwargs:
            kwargs["batch_size"] = 128

        kwargs["aggregation_length"] = aggregation_length
        kwargs["smoothing_length"] = smoothing_length

        known_mode = model_processor.is_valid_mode(prediction_mode)
        if not known_mode:
            raise ValueError(f"Error: incompatible operation mode {prediction_mode}!")

        # check if we need to use the frequency token, get token if needed
        use_frequency_token = getattr(model.config, "resolution_prefix_tuning", False)

        # needed for TTM support
        if use_frequency_token and "feature_extractor" in kwargs:
            freq = getattr(kwargs["feature_extractor"], "freq", None)
            kwargs["frequency_token"] = (
                kwargs["feature_extractor"].get_frequency_token(freq) if freq is not None else None
            )
        else:
            kwargs["frequency_token"] = None

        kwargs["mode"] = prediction_mode

        self._model_processor = model_processor
        # possibly save the posthoc_processor too?

        super().__init__(model, *args, **kwargs)

        self.__context_memory = {}
        if aggr_function.lower() == AggregationFunction.MIN.value:
            aggr_function_ = np.min
            select_function_ = np.argmin
        elif aggr_function.lower() == AggregationFunction.MEAN.value:
            aggr_function_ = np.mean
            select_function_ = None
        elif aggr_function.lower() == AggregationFunction.MAX.value:
            aggr_function_ = np.max
            select_function_ = np.argmax
        else:
            raise ValueError(
                f"Unsupported aggregation function provided {aggr_function}, expected one of {[c.value for c in AggregationFunction]}"
            )
        self.aggr_function = aggr_function_
        self.select_function = select_function_

        if self.framework == "tf":
            raise ValueError(f"The {self.__class__} is only available in PyTorch.")

    @property
    def model_type(self) -> str:
        """Returns corresponding short name of the associated model class.

        Raises:
            ValueError: unsupported model type.

        Returns:
            str: short name of the model associated with the pipeline
        """
        if isinstance(self.model, TSPulseForReconstruction):
            return "tspulse"
        elif isinstance(self.model, TinyTimeMixerForPrediction):
            return "ttm"
        raise ValueError(f"Error: unsupported model type {self.model.__class__}!")

    def _sanitize_parameters(self, **kwargs):
        """Assigns parameters to the different steps of the process. If context_length and prediction_length
        are not provided they are taken from the model config.

        For expected parameters see the call method below.
        """
        preprocess_kwargs = {}
        postprocess_kwargs = {}

        preprocess_params = [
            "context_length",
            "prediction_length",
            "timestamp_column",
            "target_columns",
            "frequency_token",
        ]
        postprocess_params = [
            "timestamp_column",
            "target_columns",
            "aggregation_length",
            "smoothing_length",
            "expand_score",
            "report_mode",
            "predictive_score_smoothing",
        ]

        for c in preprocess_params:
            if c in kwargs:
                preprocess_kwargs[c] = kwargs[c]

        for c in postprocess_params:
            if c in kwargs:
                postprocess_kwargs[c] = kwargs[c]

        if self.model_type == "tspulse":
            preprocess_kwargs["prediction_length"] = 1  # should not override model setting when TTM

        # same logic as HF Pipeline
        batch_size = kwargs.get("batch_size", self._batch_size)
        if batch_size is None:
            if self._batch_size is None:
                batch_size = 1
            else:
                batch_size = self._batch_size

        num_workers = kwargs.get("num_workers", self._num_workers)
        if num_workers is None:
            if self._num_workers is None:
                num_workers = 0
            else:
                num_workers = self._num_workers

        forward_kwargs = {
            "batch_size": batch_size,
            "num_workers": num_workers,
        }

        for p in ["mode", "aggregation_length", "expand_score"]:
            if p in kwargs:
                forward_kwargs[p] = kwargs[p]

        return preprocess_kwargs, forward_kwargs, postprocess_kwargs

    def preprocess(self, input_, **kwargs) -> Dict[str, Union[GenericTensor, List[Any]]]:
        """Preprocess step
        Load the data, if not already loaded, and then generate a pytorch dataset.
        """
        # use the feature extractor here
        if isinstance(input_, str):
            timestamp_column: str = kwargs.get("timestamp_column", "")
            if not isinstance(timestamp_column, str) or (timestamp_column == ""):
                raise ValueError("Error: timestamp column must be specified!")

            input_ = pd.read_csv(
                input_,
                parse_dates=[timestamp_column],
            )

        # For any forecasting based methods
        if self.feature_extractor:
            input_prep = self.feature_extractor.preprocess(input_)
        else:
            input_prep = input_

        # possibly calibrate posthoc_processor if calibration data is passed (not enabled yet)

        # Fixing stride to 1
        kwargs["stride"] = 1

        # maintaining processed data reference
        # CHECK if there is a preprocessor should we pass the preprocessed data here instead?
        processed_input_ = self._model_processor.preprocess(input_prep, **kwargs)

        # use forecasting dataset to do the preprocessing
        dataset = ForecastDFDataset(
            processed_input_,
            **kwargs,
        )
        self.__context_memory["data"] = input_
        self.__context_memory["reference"] = processed_input_
        return {"dataset": dataset}

    def run_single(self, inputs, preprocess_params, forward_params, postprocess_params) -> pd.DataFrame:
        """Replaces base `run_single` method which does batching during inference. This is needed to support
        large inference requests.

        Args:
            inputs (str | pd.DataFrame): data input
            preprocess_params (dict): required parameters for data preprocessing
            forward_params (dict): required parameters for model evaluation
            postprocess_params (dict): required parameters for output post processing
        Returns:
            pd.DataFrame: pipeline output
        """
        dataset = self.preprocess(inputs, **preprocess_params)["dataset"]
        signature = inspect.signature(self.model.forward)
        signature_columns = list(signature.parameters.keys())

        remove_columns_collator = RemoveColumnsCollator(
            data_collator=default_data_collator,
            signature_columns=signature_columns,
            logger=None,
            description=None,
            model_name=None,
        )

        batch_size = forward_params.get("batch_size")
        num_workers = forward_params.get("num_workers")
        dataloader = DataLoader(
            dataset, batch_size=batch_size, num_workers=num_workers, collate_fn=remove_columns_collator, shuffle=False
        )

        it = iter(dataloader)
        accumulator = defaultdict(list)

        with torch.no_grad():  # check if really needed
            while (batch := next(it, None)) is not None:
                scores = self.forward(batch, **forward_params)
                for key in scores:
                    accumulator[key].append(scores[key])

        # call postprocess
        outputs = self.postprocess(ModelOutput(accumulator), **postprocess_params)

        return outputs

    def _forward(self, input_tensors, **kwargs):
        """Forward step
        Responsible for taking pre-processed dictionary of tensors and passing it to
        the model. Aligns model parameters with the proper input parameters. Only passes
        the needed parameters from the dictionary to the model, but adds them back to the
        ouput for the next step.

        The keys in model_outputs are governed by the underlying model combined with any
        original input keys.
        """
        return self._model_processor.compute_score(input_tensors, **kwargs)

    def postprocess(self, model_outputs, **postprocess_parameters):
        """Overrides the postprocess of the base class. Applies post-processing logic on the model outputs.

        Args:
            model_outputs (dict): dictionary containing model outputs.

        Raises:
            RuntimeError: Returned if there is an inconsistency in the target columns and the resulting scores.

        Returns:
            pd.DataFrame: pandas dataframe with anomaly score attached
        """
        result = self.__context_memory["data"].copy()
        expand_score = postprocess_parameters.get("expand_score", False)
        smoothing_window_size = postprocess_parameters.get("smoothing_length", 1)
        target_columns = postprocess_parameters.get("target_columns")

        report_mode = postprocess_parameters.get("report_mode", False)
        predictive_score_smoothing = postprocess_parameters.get("predictive_score_smoothing", False)
        if not isinstance(smoothing_window_size, int):
            try:
                smoothing_window_size = int(smoothing_window_size)
            except ValueError:
                smoothing_window_size = 1

        # adjust scoring and smooth
        extra_kwargs = {}
        if "reference" in self.__context_memory:
            data = self.__context_memory["reference"]
            if len(target_columns) > 0:
                data = data[target_columns]
            extra_kwargs["reference"] = data.values

        model_outputs_ = {}
        for k in model_outputs:
            score = model_outputs[k]
            score = self._model_processor.adjust_boundary(k, score, **extra_kwargs)
            if not predictive_score_smoothing and (
                k == AnomalyScoreMethods.PREDICTIVE.value
            ):  # Skip Smoothing For 1 Lookahead forecast
                model_outputs_[k] = score
            elif k == AnomalyScoreMethods.PROBABILISTIC.value:
                model_outputs_[k] = score_smoothing(
                    score, smoothing_window_size=1
                )  # no smoothing of p-value scores across time
            else:
                model_outputs_[k] = score_smoothing(score, smoothing_window_size=smoothing_window_size)

        # aggregate scores and expand
        score = np.stack([score_ for _, score_ in model_outputs_.items()], axis=0)
        mode_selected = None
        if report_mode and (self.select_function is not None):
            keys = [key for key, _ in model_outputs_.items()]
            sel_index = self.select_function(score, axis=0)
            mode_selected = np.asarray([keys[z] for z in sel_index.ravel()]).reshape(sel_index.shape)

        score = self.aggr_function(score, axis=0)

        expand_score = (len(target_columns) > 1) and expand_score
        model_outputs = {}
        if expand_score:
            if len(target_columns) != score.shape[-1]:
                raise RuntimeError(f"Error: inconsistent state, with target columns {target_columns}")
            for i, col_name in enumerate(target_columns):
                model_outputs[f"{col_name}_anomaly_score"] = score[..., i]

            if mode_selected is not None:
                for i, col_name in enumerate(target_columns):
                    model_outputs[f"{col_name}_selected_mode"] = mode_selected[..., i]
        else:
            model_outputs.update(anomaly_score=score.ravel())
            if mode_selected is not None:
                model_outputs.update(selected_mode=mode_selected.ravel())

        # populate dataframe
        for k in model_outputs:
            result[k] = model_outputs[k]
        self.__context_memory = {}
        return result



# Standard
import json
import sys
import traceback

import torch

from tsfmfinetuning.error_logging import write_termination_log
from tsfmfinetuning.finetuning import FinetuningRuntime
from tsfmfinetuning.ftargs import argparser
from tsfmfinetuning.ftpayloads import TinyTimeMixerForecastingTuneInput


# remote container space
def main():
    # ###########################3
    args, _ = argparser().parse_known_args()

    # reconstruct our input object
    # these come from a mutually exclusive group
    payload: dict = json.load(open(args.payload, "r"))
    # this will give us param validation

    if args.model_arch == "ttm" and args.task_type == "forecasting":
        input: TinyTimeMixerForecastingTuneInput = TinyTimeMixerForecastingTuneInput(**payload)
        ftr: FinetuningRuntime = FinetuningRuntime()
        ftr.finetuning(input=input, tuned_model_name=args.model_name, output_dir=args.target_dir)
    else:
        raise NotImplementedError(f"model arch/task type not implemented {args.model_arch_type} {args.task_type}")


if __name__ == "__main__":
    try:
        if torch.cuda.is_available() and torch.cuda.device_count() >= 1:
            print(f"{torch.cuda.device_count()} gpu device(s) are available for use.")
        else:
            print("no gpu or torch.cuda is not available, using CPU")
        main()
        sys.exit(0)
    except Exception as e:
        traceback.print_exception(e)
        write_termination_log(f"Exception when running finetuning {e}")
        sys.exit(1)



# Copyright contributors to the TSFM project
#
"""Utilities to support services"""

from pathlib import Path
from typing import Optional, Union

from transformers import PreTrainedModel

from .time_series_preprocessor import TimeSeriesPreprocessor
from .tsfm_config import TSFMConfig


# tuples of:
# "inference_handler_module_path", "inference_handler_class_name",  "tuning_handler_module_path",  "tuning_handler_class_name")

service_handler_mapping = {
    "TinyTimeMixerConfig": (
        "tsfminference.tsfm_inference_handler",
        "TinyTimeMixerForecastingInferenceHandler",
        "tsfmfinetuning.tsfm_tuning_handler",
        "TinyTimeMixerForecastingTuningHandler",
    ),
    "PatchTSTConfig": (
        "tsfminference.tsfm_inference_handler",
        "ForecastingInferenceHandler",
        "tsfmfinetuning.tsfm_tuning_handler",
        "ForecastingTuningHandler",
    ),
    "PatchTSMixerConfig": (
        "tsfminference.tsfm_inference_handler",
        "ForecastingInferenceHandler",
        "tsfmfinetuning.tsfm_tuning_handler",
        "ForecastingTuningHandler",
    ),
}


def save_deployment_package(
    save_path: Union[Path, str],
    model: PreTrainedModel,
    ts_processor: Optional[TimeSeriesPreprocessor] = None,
    **kwargs,
):
    """Convenience function for saving the deployment package needed to use the services.

    Args:
        save_path (Union[Path, str]): Path to location to save files, a folder will be created at this path.
        model (PreTrainedModel): The model for which you wish to create the deployment package.
        ts_processor (Optional[TimeSeriesProcessor], optional): The time series processor used when training or
            finetuning the model. Defaults to None.

        Currently supports TinyTimeMixer, PatchTSMixer, PatchTST. Supported models need to have an entry in
        service_handler_mapping.

    Raises:
        ValueError: Raised if model is not one of the supported types.
    """
    if isinstance(save_path, str):
        save_path = Path(save_path)

    model.save_pretrained(save_path)

    if ts_processor is not None:
        ts_processor.save_pretrained(save_path)

    # now create and save tsfm config
    params = {}

    config_class = model.config.__class__.__name__
    handler_params = service_handler_mapping.get(config_class, None)

    if handler_params is None:
        raise ValueError(f"Could not find suitable handler information for config class {config_class}")

    (
        params["inference_handler_module_path"],
        params["inference_handler_class_name"],
        params["tuning_handler_module_path"],
        params["tuning_handler_class_name"],
    ) = handler_params

    params["model_config_name"] = config_class
    params["model_class_name"] = model.__class__.__name__
    params["model_type"] = model.config.model_type
    params["module_path"] = model.__class__.__module__  # "tsfm_public", maybe filter

    params["is_finetuned"] = ts_processor is not None

    # assumes we are dealing with one of the known IBM models
    params["minimum_context_length"] = kwargs.pop("minimum_context_length", model.config.context_length)
    params["maximum_context_length"] = kwargs.pop("maximum_context_length", model.config.context_length)
    params["maximum_prediction_length"] = kwargs.pop("maximum_prediction_length", model.config.prediction_length)
    params.update(**kwargs)

    svc_config = TSFMConfig(**params)

    svc_config.save_pretrained(save_path)



# Copyright contributors to the TSFM project
#

"""Tests the base processor class"""

import tempfile

from tsfm_public.toolkit.processor import BaseProcessor


def test_processor_saves_with_custom_name():
    class MyProcessor(BaseProcessor):
        PROCESSOR_NAME = "my_processor_config.json"

    p = MyProcessor(my_param="10")

    with tempfile.TemporaryDirectory() as d:
        p.save_pretrained(d)
        p_new = MyProcessor.from_pretrained(d)

        assert p.my_param == p_new.my_param



"""A configuration class for service handlers"""
# Adapted from https://github.com/huggingface/transformers/blob/main/src/transformers/configuration_utils.py

import copy
import json
import logging
import os
import warnings
from typing import Any, Dict, Optional, Tuple, Union

from transformers.utils import PushToHubMixin, cached_file, download_url, is_remote_url


TSFM_CONFIG_NAME = "tsfm_config.json"

LOGGER = logging.getLogger(__file__)


class TSFMConfig(PushToHubMixin):
    """A configuration class used by the time series serving framework.

    Args:
        inference_handler_module_path (str, optional): module path where the inference handler class is located.
        inference_handler_class_name (str, optional): class name for the inference handler. If ommitted, the default
            ForecastingHuggingFaceInferenceHandler will be used.
        tuning_handler_module_path (str, optional): module path where the tuning handler class is located.
        tuning_handler_class_name (str, optional): class name for the tuning handler. If ommitted, the default
            ForecastingHuggingFaceTuningHandler will be used.
        module_path (str, optional): module path where the model class is located.
        model_type (str, optional): An identifier for the model type, serialized into the model config JSON file, and used
            to recreate the correct object in AutoConfig. For this config, it should match the model_type of the model that
            will be served.
        model_config_name (str, optional): the name of the model configuration class.
        model_class_name (str, optional): the class name of the model.
        exogenous_support (bool, optional): True if the model supports exogenous covariates. Defaults to False.
        multivariate_support (bool, optional): True if the model supports multivariate inputs/outputs. Defaults to True.
        missing_value_support (bool, optional): True if hte model supports missing values in the input. Defaults to False.
        minimum_context_length (int, optional): Minimum context length required by the model. Defaults to 1.
        maximum_context_length (int, optional): Maximum input context length consumable by the model. Use None to indicate
            no maximum. Defaults to None.
        maximum_prediction_length (int, optional): Maximum prediction length produced by the model. Use None to indicate
            no maximum. Defaults to None.
        inference_batch_size (int, optional): Size of batch to construct when performing inference. Defaults to 16.
        is_finetuned (bool, optional): Indicates that the model has been finetuned. Defaults to False.
        extra_pipeline_arguments (dict, optional): Additional arguments that get passed during service operations. Defaults to {}.

    Notes:
        If module_path, model_type, model_config_name, model_class_name are absent, and the HFServiceHandler is used, the
        model will be loaded using standard AutoConfig/AutoModel functions from HF.
    """

    def __init__(self, **kwargs):
        # Attributes with defaults
        self.inference_handler_module_path = kwargs.pop("inference_handler_module_path", None)
        self.inference_handler_class_name = kwargs.pop("inference_handler_class_name", None)
        self.tuning_handler_module_path = kwargs.pop("tuning_handler_module_path", None)
        self.tuning_handler_class_name = kwargs.pop("tuning_handler_class_name", None)
        self.module_path = kwargs.pop("module_path", None)
        self.model_type = kwargs.pop("model_type", None)
        self.model_config_name = kwargs.pop("model_config_name", None)
        self.model_class_name = kwargs.pop("model_class_name", None)
        self.exogenous_support = kwargs.pop("exogenous_support", False)
        self.multivariate_support = kwargs.pop("multivariate_support", True)
        self.missing_value_support = kwargs.pop("missing_value_support", False)
        self.minimum_context_length = kwargs.pop("minimum_context_length", 1)
        self.maximum_context_length = kwargs.pop("maximum_context_length", None)
        self.maximum_prediction_length = kwargs.pop("maximum_prediction_length", None)
        self.is_finetuned = kwargs.pop("is_finetuned", False)
        self.inference_batch_size = kwargs.pop("inference_batch_size", 16)
        self.extra_pipeline_arguments = kwargs.pop("extra_pipeline_arguments", {})

        # "maximum_prediction_length": 96,
        # "minimum_context_length": 512,
        # "maximum_context_length": 512,
        # "model_type": "tinytimemixer",
        # "model_config_name": "TinyTimeMixerConfig",
        # "model_class_name": "TinyTimeMixerForPrediction"

        # Name or path to the config
        self._name_or_path = str(kwargs.pop("name_or_path", ""))
        # Config hash
        # self._commit_hash = kwargs.pop("_commit_hash", None)

        # Additional attributes without default values
        for key, value in kwargs.items():
            try:
                setattr(self, key, value)
            except AttributeError as err:
                LOGGER.error(f"Can't set {key} with value {value} for {self}")
                raise err

    @property
    def name_or_path(self) -> str:
        return getattr(self, "_name_or_path", None)

    @name_or_path.setter
    def name_or_path(self, value):
        self._name_or_path = str(value)  # Make sure that name_or_path is a string (for JSON encoding)

    @staticmethod
    def _set_token_in_kwargs(kwargs, token=None):
        """Temporary method to deal with `token` and `use_auth_token`.

        This method is to avoid apply the same changes in all model config classes that overwrite `from_pretrained`.

        Need to clean up `use_auth_token` in a follow PR.
        """
        # Some model config classes like CLIP define their own `from_pretrained` without the new argument `token` yet.
        if token is None:
            token = kwargs.pop("token", None)
        use_auth_token = kwargs.pop("use_auth_token", None)

        if use_auth_token is not None:
            warnings.warn(
                "The `use_auth_token` argument is deprecated and will be removed in v5 of Transformers. Please use `token` instead.",
                FutureWarning,
            )
            if token is not None:
                raise ValueError(
                    "`token` and `use_auth_token` are both specified. Please set only the argument `token`."
                )
            token = use_auth_token

        if token is not None:
            kwargs["token"] = token

    def save_pretrained(self, save_directory: Union[str, os.PathLike], push_to_hub: bool = False, **kwargs):
        """
        Save a configuration object to the directory `save_directory`, so that it can be re-loaded using the
        [`~PretrainedConfig.from_pretrained`] class method.

        Args:
            save_directory (`str` or `os.PathLike`):
                Directory where the configuration JSON file will be saved (will be created if it does not exist).
            push_to_hub (`bool`, *optional*, defaults to `False`):
                Whether or not to push your model to the Hugging Face model hub after saving it. You can specify the
                repository you want to push to with `repo_id` (will default to the name of `save_directory` in your
                namespace).
            kwargs (`Dict[str, Any]`, *optional*):
                Additional key word arguments passed along to the [`~utils.PushToHubMixin.push_to_hub`] method.
        """
        self._set_token_in_kwargs(kwargs)

        if os.path.isfile(save_directory):
            raise AssertionError(f"Provided path ({save_directory}) should be a directory, not a file")

        os.makedirs(save_directory, exist_ok=True)

        if push_to_hub:
            commit_message = kwargs.pop("commit_message", None)
            repo_id = kwargs.pop("repo_id", save_directory.split(os.path.sep)[-1])
            repo_id = self._create_repo(repo_id, **kwargs)
            files_timestamps = self._get_files_timestamps(save_directory)

        # If we have a custom config, we copy the file defining it in the folder and set the attributes so it can be
        # loaded from the Hub.
        # if self._auto_class is not None:
        #     custom_object_save(self, save_directory, config=self)

        # If we save using the predefined names, we can load using `from_pretrained`
        output_config_file = os.path.join(save_directory, TSFM_CONFIG_NAME)

        self.to_json_file(output_config_file, use_diff=True)
        LOGGER.info(f"Configuration saved in {output_config_file}")

        if push_to_hub:
            self._upload_modified_files(
                save_directory,
                repo_id,
                files_timestamps,
                commit_message=commit_message,
                token=kwargs.get("token"),
            )

    def to_json_file(self, json_file_path: Union[str, os.PathLike], use_diff: bool = True):
        """
        Save this instance to a JSON file.

        Args:
            json_file_path (`str` or `os.PathLike`):
                Path to the JSON file in which this configuration instance's parameters will be saved.
            use_diff (`bool`, *optional*, defaults to `True`):
                If set to `True`, only the difference between the config instance and the default `PretrainedConfig()`
                is serialized to JSON file.
        """
        with open(json_file_path, "w", encoding="utf-8") as writer:
            writer.write(self.to_json_string(use_diff=use_diff))

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: Union[str, os.PathLike],
        cache_dir: Optional[Union[str, os.PathLike]] = None,
        force_download: bool = False,
        local_files_only: bool = False,
        token: Optional[Union[str, bool]] = None,
        revision: str = "main",
        **kwargs,
    ) -> "TSFMConfig":
        r"""
        Instantiate a [`PretrainedConfig`] (or a derived class) from a pretrained model configuration.

        Args:
            pretrained_model_name_or_path (`str` or `os.PathLike`):
                This can be either:

                - a string, the *model id* of a pretrained model configuration hosted inside a model repo on
                  huggingface.co.
                - a path to a *directory* containing a configuration file saved using the
                  [`~PretrainedConfig.save_pretrained`] method, e.g., `./my_model_directory/`.
                - a path or url to a saved configuration JSON *file*, e.g., `./my_model_directory/configuration.json`.
            cache_dir (`str` or `os.PathLike`, *optional*):
                Path to a directory in which a downloaded pretrained model configuration should be cached if the
                standard cache should not be used.
            force_download (`bool`, *optional*, defaults to `False`):
                Whether or not to force to (re-)download the configuration files and override the cached versions if
                they exist.
            resume_download:
                Deprecated and ignored. All downloads are now resumed by default when possible.
                Will be removed in v5 of Transformers.
            proxies (`Dict[str, str]`, *optional*):
                A dictionary of proxy servers to use by protocol or endpoint, e.g., `{'http': 'foo.bar:3128',
                'http://hostname': 'foo.bar:4012'}.` The proxies are used on each request.
            token (`str` or `bool`, *optional*):
                The token to use as HTTP bearer authorization for remote files. If `True`, or not specified, will use
                the token generated when running `huggingface-cli login` (stored in `~/.huggingface`).
            revision (`str`, *optional*, defaults to `"main"`):
                The specific model version to use. It can be a branch name, a tag name, or a commit id, since we use a
                git-based system for storing models and other artifacts on huggingface.co, so `revision` can be any
                identifier allowed by git.

                <Tip>

                To test a pull request you made on the Hub, you can pass `revision="refs/pr/<pr_number>"`.

                </Tip>

            return_unused_kwargs (`bool`, *optional*, defaults to `False`):
                If `False`, then this function returns just the final configuration object.

                If `True`, then this functions returns a `Tuple(config, unused_kwargs)` where *unused_kwargs* is a
                dictionary consisting of the key/value pairs whose keys are not configuration attributes: i.e., the
                part of `kwargs` which has not been used to update `config` and is otherwise ignored.
            subfolder (`str`, *optional*, defaults to `""`):
                In case the relevant files are located inside a subfolder of the model repo on huggingface.co, you can
                specify the folder name here.
            kwargs (`Dict[str, Any]`, *optional*):
                The values in kwargs of any keys which are configuration attributes will be used to override the loaded
                values. Behavior concerning key/value pairs whose keys are *not* configuration attributes is controlled
                by the `return_unused_kwargs` keyword parameter.

        Returns:
            [`PretrainedConfig`]: The configuration object instantiated from this pretrained model.

        Examples:

        ```python
        # We can't instantiate directly the base class *PretrainedConfig* so let's show the examples on a
        # derived class: BertConfig
        config = BertConfig.from_pretrained(
            "google-bert/bert-base-uncased"
        )  # Download configuration from huggingface.co and cache.
        config = BertConfig.from_pretrained(
            "./test/saved_model/"
        )  # E.g. config (or model) was saved using *save_pretrained('./test/saved_model/')*
        config = BertConfig.from_pretrained("./test/saved_model/my_configuration.json")
        config = BertConfig.from_pretrained("google-bert/bert-base-uncased", output_attentions=True, foo=False)
        assert config.output_attentions == True
        config, unused_kwargs = BertConfig.from_pretrained(
            "google-bert/bert-base-uncased", output_attentions=True, foo=False, return_unused_kwargs=True
        )
        assert config.output_attentions == True
        assert unused_kwargs == {"foo": False}
        ```"""
        kwargs["cache_dir"] = cache_dir
        kwargs["force_download"] = force_download
        kwargs["local_files_only"] = local_files_only
        kwargs["revision"] = revision

        cls._set_token_in_kwargs(kwargs, token)

        config_dict, kwargs = cls.get_config_dict(pretrained_model_name_or_path, **kwargs)

        return cls.from_dict(config_dict, **kwargs)

    @classmethod
    def get_config_dict(
        cls, pretrained_model_name_or_path: Union[str, os.PathLike], **kwargs
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        From a `pretrained_model_name_or_path`, resolve to a dictionary of parameters, to be used for instantiating a
        [`PretrainedConfig`] using `from_dict`.

        Parameters:
            pretrained_model_name_or_path (`str` or `os.PathLike`):
                The identifier of the pre-trained checkpoint from which we want the dictionary of parameters.

        Returns:
            `Tuple[Dict, Dict]`: The dictionary(ies) that will be used to instantiate the configuration object.

        """
        cls._set_token_in_kwargs(kwargs)

        # Get config dict associated with the base config file
        config_dict, kwargs = cls._get_config_dict(pretrained_model_name_or_path, **kwargs)
        if config_dict is None:
            return {}, kwargs

        return config_dict, kwargs

    @classmethod
    def _get_config_dict(
        cls, pretrained_model_name_or_path: Union[str, os.PathLike], **kwargs
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        cache_dir = kwargs.pop("cache_dir", None)
        force_download = kwargs.pop("force_download", False)
        resume_download = kwargs.pop("resume_download", None)
        proxies = kwargs.pop("proxies", None)
        token = kwargs.pop("token", None)
        local_files_only = kwargs.pop("local_files_only", False)
        revision = kwargs.pop("revision", None)
        trust_remote_code = kwargs.pop("trust_remote_code", None)
        subfolder = kwargs.pop("subfolder", "")
        from_pipeline = kwargs.pop("_from_pipeline", None)
        from_auto_class = kwargs.pop("_from_auto", False)

        if trust_remote_code is True:
            LOGGER.warning(
                "The argument `trust_remote_code` is to be used with Auto classes. It has no effect here and is"
                " ignored."
            )

        user_agent = {"file_type": "config", "from_auto_class": from_auto_class}
        if from_pipeline is not None:
            user_agent["using_pipeline"] = from_pipeline

        pretrained_model_name_or_path = str(pretrained_model_name_or_path)

        is_local = os.path.isdir(pretrained_model_name_or_path)
        if os.path.isfile(os.path.join(subfolder, pretrained_model_name_or_path)):
            # Special case when pretrained_model_name_or_path is a local file
            resolved_config_file = pretrained_model_name_or_path
            is_local = True
        elif is_remote_url(pretrained_model_name_or_path):
            configuration_file = pretrained_model_name_or_path
            resolved_config_file = download_url(pretrained_model_name_or_path)
        else:
            configuration_file = kwargs.pop("_configuration_file", TSFM_CONFIG_NAME)

            try:
                # Load from local folder or from cache or download from model Hub and cache
                resolved_config_file = cached_file(
                    pretrained_model_name_or_path,
                    configuration_file,
                    cache_dir=cache_dir,
                    force_download=force_download,
                    proxies=proxies,
                    resume_download=resume_download,
                    local_files_only=local_files_only,
                    token=token,
                    user_agent=user_agent,
                    revision=revision,
                    subfolder=subfolder,
                )
                if resolved_config_file is None:
                    return None, kwargs

            except EnvironmentError:
                # Raise any environment error raise by `cached_file`. It will have a helpful error message adapted to
                # the original exception.
                raise
            except Exception:
                # For any other exception, we throw a generic error.
                raise EnvironmentError(
                    f"Can't load the configuration of '{pretrained_model_name_or_path}'. If you were trying to load it"
                    " from 'https://huggingface.co/models', make sure you don't have a local directory with the same"
                    f" name. Otherwise, make sure '{pretrained_model_name_or_path}' is the correct path to a directory"
                    f" containing a {configuration_file} file"
                )

        try:
            config_dict = cls._dict_from_json_file(resolved_config_file)
        except (json.JSONDecodeError, UnicodeDecodeError):
            raise EnvironmentError(
                f"It looks like the config file at '{resolved_config_file}' is not a valid JSON file."
            )

        if is_local:
            LOGGER.info(f"loading configuration file {resolved_config_file}")
        else:
            LOGGER.info(f"loading configuration file {configuration_file} from cache at {resolved_config_file}")

        return config_dict, kwargs

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any], **kwargs) -> "TSFMConfig":
        """
        Instantiates a [`PretrainedConfig`] from a Python dictionary of parameters.

        Args:
            config_dict (`Dict[str, Any]`):
                Dictionary that will be used to instantiate the configuration object. Such a dictionary can be
                retrieved from a pretrained checkpoint by leveraging the [`~PretrainedConfig.get_config_dict`] method.
            kwargs (`Dict[str, Any]`):
                Additional parameters from which to initialize the configuration object.

        Returns:
            [`PretrainedConfig`]: The configuration object instantiated from those parameters.
        """
        return_unused_kwargs = kwargs.pop("return_unused_kwargs", False)
        # Those arguments may be passed along for our internal telemetry.
        # We remove them so they don't appear in `return_unused_kwargs`.
        # The commit hash might have been updated in the `config_dict`, we don't want the kwargs to erase that update.

        config = cls(**config_dict)

        to_remove = []
        for key, value in kwargs.items():
            if hasattr(config, key):
                current_attr = getattr(config, key)
                # To authorize passing a custom subconfig as kwarg in models that have nested configs.
                if isinstance(current_attr, TSFMConfig) and isinstance(value, dict):
                    value = current_attr.__class__(**value)
                setattr(config, key, value)
                to_remove.append(key)
        for key in to_remove:
            kwargs.pop(key, None)

        LOGGER.info(f"TSFM config {config}")
        if return_unused_kwargs:
            return config, kwargs
        else:
            return config

    @classmethod
    def from_json_file(cls, json_file: Union[str, os.PathLike]) -> "TSFMConfig":
        """
        Instantiates a [`PretrainedConfig`] from the path to a JSON file of parameters.

        Args:
            json_file (`str` or `os.PathLike`):
                Path to the JSON file containing the parameters.

        Returns:
            [`PretrainedConfig`]: The configuration object instantiated from that JSON file.

        """
        config_dict = cls._dict_from_json_file(json_file)
        return cls(**config_dict)

    @classmethod
    def _dict_from_json_file(cls, json_file: Union[str, os.PathLike]):
        with open(json_file, "r", encoding="utf-8") as reader:
            text = reader.read()
        return json.loads(text)

    def __eq__(self, other):
        return isinstance(other, TSFMConfig) and (self.__dict__ == other.__dict__)

    def __repr__(self):
        return f"{self.__class__.__name__} {self.to_json_string()}"

    def __iter__(self):
        for attr in self.__dict__:
            yield attr

    def to_diff_dict(self) -> Dict[str, Any]:
        """
        Removes all attributes from config which correspond to the default config attributes for better readability and
        serializes to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this configuration instance,
        """
        config_dict = self.to_dict()

        # get the default config dict
        default_config_dict = TSFMConfig().to_dict()

        # get class specific config dict
        class_config_dict = self.__class__().to_dict()

        serializable_config_dict = {}

        # only serialize values that differ from the default config
        for key, value in config_dict.items():
            if (
                key not in default_config_dict
                or value != default_config_dict[key]
                or (key in class_config_dict and value != class_config_dict[key])
            ):
                serializable_config_dict[key] = value

        return serializable_config_dict

    def to_dict(self) -> Dict[str, Any]:
        """
        Serializes this instance to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this configuration instance.
        """
        output = copy.deepcopy(self.__dict__)

        for key, value in output.items():
            # Deal with nested configs like CLIP
            if isinstance(value, TSFMConfig):
                value = value.to_dict()

            output[key] = value

        return output

    def to_json_string(self, use_diff: bool = True) -> str:
        """
        Serializes this instance to a JSON string.

        Args:
            use_diff (`bool`, *optional*, defaults to `True`):
                If set to `True`, only the difference between the config instance and the default `PretrainedConfig()`
                is serialized to JSON string.

        Returns:
            `str`: String containing all the attributes that make up this configuration instance in JSON format.
        """
        if use_diff is True:
            config_dict = self.to_diff_dict()
        else:
            config_dict = self.to_dict()
        return json.dumps(config_dict, indent=2, sort_keys=True) + "\n"



import base64
import datetime
import io
import tempfile

import numpy as np
import pandas as pd
import pytest
from tsfmfinetuning.ioutils import _split_text_file, to_pandas


def test_to_pandas_with_feather_content_from_filesystem():
    df = pd.read_feather("data/ETTh1.feather")
    df2 = to_pandas(uri="file://./data/ETTh1.feather")
    assert isinstance(df2.dtypes["date"], np.dtypes.DateTime64DType)
    assert df.equals(df2)


def test_to_pandas_with_csv_content_from_filesystem():
    df = pd.read_csv(filepath_or_buffer="data/ETTh1.csv", parse_dates=["date"])
    df2 = to_pandas(uri="file://./data/ETTh1.csv", timestamp_column="date")
    assert isinstance(df2.dtypes["date"], np.dtypes.DateTime64DType)
    assert df.equals(df2)


def test_to_pandas_with_multipart_csv_content_from_filesystem():
    df = pd.read_csv(filepath_or_buffer="data/ETTh1.csv", parse_dates=["date"])
    df2 = to_pandas(uri="file://./data/multipart")
    assert isinstance(df2.dtypes["date"], np.dtypes.DateTime64DType)
    # we can't check equality as pyarrow is parsing dates like '2016-07-01 00:00:00' to second
    # resolution whereas pd.read_csv resolves it to ns resolution. I think pyparrow is "more" correct.
    assert len(df) == len(df2)
    assert list(df.columns) == list(df2.columns)
    np.array_equal(df.to_numpy(), df2.to_numpy())


def test___split_text_file():
    original = open("./data/ETTh1.csv").readlines()
    with tempfile.TemporaryDirectory() as td:
        parts = 10
        filenames = _split_text_file(
            source="./data/ETTh1.csv", target_dir=td, has_header=True, shuffle=False, parts=parts
        )
        sum = 0
        for fn in filenames:
            sum += len(open(fn).readlines())
        sum -= parts - 1  # account for added header in each part
        assert len(original) == sum, "Line lengths do not match"
        assert original[-1] == open(filenames[-1]).readlines()[-1]
        assert original[1] == open(filenames[0]).readlines()[1]


def test_to_pandas_with_unsupported_uri():
    with pytest.raises(NotImplementedError) as _:
        to_pandas("https://foobaz/fizbot.csv")
    with pytest.raises(NotImplementedError) as _:
        to_pandas("s3a://tsfm-services/ETTh1.feather")
    with pytest.raises(NotImplementedError) as _:
        to_pandas("http://tsfm-services/ETTh1.feather")
    with pytest.raises(NotImplementedError) as _:
        to_pandas("./data/ETTh1.csv")  # we require a proper uri
    with pytest.raises(ValueError) as _:
        to_pandas("./data/foo.csv")
    with pytest.raises(ValueError) as _:
        to_pandas("./data/foo")  # passing multipart that's not there


def test_to_pandas_with_bad_binary_content():
    with pytest.raises(Exception):
        to_pandas(uri="makesnosense")


def test_to_pandas_with_base64_feathered_content():
    df = pd.DataFrame(columns=["date", "value"], data=[[datetime.datetime.now(), 1.0]])
    buf = io.BytesIO()
    df.to_feather(buf)
    buf.seek(0)
    output = base64.b64encode(buf.getbuffer())
    df2 = to_pandas(uri=output.decode("utf-8"), timestamp_column="date")
    assert df.equals(df2)


def test_to_pandas_with_compressed_csv_from_filesystem():
    df2 = to_pandas(uri="file://./data/ETTh1.csv.gz", timestamp_column="date")
    assert isinstance(df2.dtypes["date"], np.dtypes.DateTime64DType)


def test_to_pandas_with_base64_csv_content():
    df = pd.DataFrame(columns=["date", "value"], data=[[datetime.datetime.now(), 1.0]])
    buf = io.BytesIO()
    df.to_csv(buf, header=True, index=False)
    buf.seek(0)
    output = base64.b64encode(buf.getbuffer())
    df2 = to_pandas(uri=output.decode("utf-8"), timestamp_column="date")
    assert df.equals(df2)


def test_to_pandas_with_unsupported_binary_content():
    df = pd.DataFrame(columns=["date", "value"], data=[[datetime.datetime.now(), 1.0]])
    buf = io.BytesIO()
    df.to_pickle(buf)
    buf.seek(0)
    output = base64.b64encode(buf.getbuffer())

    with pytest.raises(Exception):
        to_pandas(uri=base64.b64encode(output.decode("utf-8")))



"""Utilities for directory operations."""

import os
from pathlib import Path


def resolve_model_path(search_path: str, model_id: str) -> Path:
    """Find the first path under search_path for model_id. All entries in
    search_path must be:
    * an existing directory
    * must be readable by the current process

    Args:
        search_path (str): A unix-like ":" separated list of directories such a "dir1:dir2"
        model_id (str): a model_id (which is really just a subdirectory under dir1 or dir2)

    Returns:
        Path: the first matching path, None if no path is fount.
    """

    _amodeldir_found = next(
        (
            adir
            for adir in (Path(p) for p in search_path.split(":"))
            if adir.exists()
            and adir.is_dir()
            and os.access(adir, os.R_OK)
            and (adir / model_id).exists()
            and os.access(adir / model_id, os.R_OK)
        ),
        None,
    )
    if not _amodeldir_found:
        return None
    return _amodeldir_found / model_id
